(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Ceta2
imports Ceta Efficient_RPO
begin

export_code certify_proof in Haskell module_name Ceta file "../generated/tests/inefficient"

declare rpo_efficient_1_sound[code]

export_code certify_proof in Haskell module_name Ceta file "../generated/tests/eff_1"

declare rpo_efficient_2_sound[code]

export_code certify_proof in Haskell module_name Ceta file "../generated/tests/eff_2"

definition approx :: "('f, 'v)term \<Rightarrow> nat" where
 "approx t = size t"

definition comp :: "nat \<Rightarrow> nat \<Rightarrow> comparison" where
  "comp m n = (if m = n then Memories.Equal else (if m > n then Memories.More else Memories.Less))"

lemma comp_eq: "(comp h i = Memories.Equal) = (h = i)"
proof (cases "h = i")
case True thus ?thesis by (simp add: comp_def)
next
case False thus ?thesis by (cases "h > i") (auto simp: comp_def)
qed

lemma comp_lt: "(comp h i = Memories.Less) = (h < i)"
proof (cases "h = i")
case True thus ?thesis by (simp add: comp_def)
next
case False thus ?thesis by (cases "h > i") (auto simp: comp_def)
qed

lemma comp_gt: "(comp h i = Memories.More) = (h > i)"
proof (cases "h = i")
case True thus ?thesis by (simp add: comp_def)
next
case False thus ?thesis by (cases "h > i") (auto simp: comp_def)
qed

lemma comp_lt_gt: "(comp h i = Memories.Less) = (comp i h = Memories.More)"
using comp_lt comp_gt by simp

definition approx_t where "approx_t = atm l2m (\<lambda>p. approx (snd p)) comp 7"
definition approx_s where "approx_s = atm approx_t (\<lambda>p. approx (fst p)) comp 7"

definition efficient_rpo_3 where
  [code del]: "efficient_rpo_3 = memoized_rpo.memoized_rpo approx_s"

interpretation approx_t: approximation_tree l2m l2m_valid "\<lambda>p. approx (snd p)" comp 7
using satisfactory_l2m comp_eq comp_lt_gt by unfold_locales auto

interpretation approx_s: approximation_tree approx_t "\<lambda>_. atm_valid l2m l2m_valid" "\<lambda>p. approx (fst p)" comp 7
using approx_t.satisfactory_atm comp_eq comp_lt_gt
by unfold_locales (simp_all add: approx_t.atm_inst_def approx_t_def)

interpretation memoized_rpo_3 : memoized_rpo approx_s "\<lambda>_. atm_valid approx_t (\<lambda>_. atm_valid l2m l2m_valid)"
  where "memoized_rpo.memoized_rpo approx_s = efficient_rpo_3"
using approx_s.satisfactory_atm unfolding approx_s.atm_inst_def
by unfold_locales (auto simp: efficient_rpo_3_def approx_s_def)

theorem rpo_efficient_3_sound: "rpo_unbounded_impl = efficient_rpo_3"
by (simp add: memoized_rpo_3.memoized_rpo_sound efficient_rpo_3_def)

declare rpo_efficient_3_sound[code]

export_code certify_proof in Haskell module_name Ceta file "../generated/tests/eff_3"


end