(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Proof_Checker
imports
  Check_Termination
  Check_Nontermination
  Check_CRP
  Check_Completion_Proof
  Check_Equational_Proof
  Check_Complexity
  Check_Quasi_Reductive
  Context_Sensitive_Impl
  Outermost_Forbidden_Patterns
begin

datatype ('f, 'v) strategy =
  No_Strategy
| Innermost
| Innermost_Q "('f, 'v) term list"

fun strategy_to_Q :: "('f, 'v) strategy \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) term list" where
  "strategy_to_Q No_Strategy _ = []"
| "strategy_to_Q Innermost R = remdups (map fst R)"
| "strategy_to_Q (Innermost_Q Q) _ = Q"

fun rel_rules_of :: "('f,'v)rules option \<Rightarrow> ('f,'v)rules"
 where "rel_rules_of None = []"
     | "rel_rules_of (Some r) = r"

datatype ('f, 'v) input_problem =
  TRS bool "('f, 'v) strategy" "('f, 'v) rules" "('f, 'v) rules"
| OTRS "('f, 'v) rules"
| CSTRS "'f muL" "('f, 'v) rules"
| FPTRS "('f, 'v) forb_pattern list" "('f, 'v) rules"
| DP bool bool "('f, 'v) rules" "('f, 'v) rules" "('f, 'v) strategy" "('f, 'v) rules" "('f, 'v) rules"
| EQS_TRS "('f, 'v) equation list" "('f, 'v) rules"
| EQS_EQ "('f, 'v) equation list" "('f, 'v) term \<times> ('f, 'v) term"
| Complexity_TRS "('f, 'v) strategy" "('f, 'v) rules" "('f, 'v) rules" "('f, 'v) complexity_measure" complexity_class
| CTRS "('f, 'v) crules"
| Unknown unknown_info

datatype ('f, 'l, 'v) cert_problem = 
  TRS_Termination_Proof bool "(('f, 'l) lab, 'v) strategy"
            "(('f, 'l) lab, 'v) rules"
            "(('f, 'l) lab, 'v) rules option"
            "('f, 'l, 'v) trs_termination_proof"
| Complexity_Proof "(('f, 'l) lab, 'v) strategy"
            "(('f, 'l) lab, 'v) rules"
            "(('f, 'l) lab, 'v) rules option"
            "(('f,'l)lab,'v)complexity_measure" complexity_class
            "(('f, 'l) lab, 'v) complexity_proof"
| DP_Termination_Proof bool bool "(('f, 'l) lab, 'v) rules"
  "(('f, 'l) lab, 'v) rules"                                    
  "(('f, 'l) lab, 'v) strategy"
  "(('f, 'l) lab, 'v) rules"
  "(('f, 'l) lab, 'v) rules"
  "('f, 'l, 'v) dp_termination_proof"
| DP_Nontermination_Proof bool bool "(('f, 'l) lab, 'v) rules"
  "(('f, 'l) lab, 'v) strategy"
  "(('f, 'l) lab, 'v) rules"
  "('f, 'l, 'v) dp_nontermination_proof"
| TRS_Nontermination_Proof bool "(('f, 'l) lab, 'v) strategy"
               "(('f, 'l) lab, 'v) rules"
               "('f, 'l, 'v) trs_nontermination_proof"
| Outermost_Termination_Proof 
               "(('f, 'l) lab, 'v) rules"
               "('f,'l,'v) fptrs_termination_proof"
| Outermost_Nontermination_Proof 
               "(('f, 'l) lab, 'v) rules"
               "('f,'l,'v) fp_nontermination_proof"
| CS_Termination_Proof "('f,'l)lab muL"
               "(('f, 'l) lab, 'v) rules"
               "('f,'l,'v) fptrs_termination_proof"
| CS_Nontermination_Proof "('f,'l)lab muL"
               "(('f, 'l) lab, 'v) rules"
               "('f,'l,'v) fp_nontermination_proof"
| FP_Termination_Proof "(('f, 'l) lab, 'v) forb_pattern list"
               "(('f, 'l) lab, 'v) rules"
               "('f, 'l, 'v) fptrs_termination_proof"
| FP_Nontermination_Proof "(('f, 'l) lab, 'v) forb_pattern list"
               "(('f, 'l) lab, 'v) rules"
               "('f,'l,'v) fp_nontermination_proof"
| Relative_TRS_Nontermination_Proof bool "(('f, 'l) lab, 'v) strategy"
                   "(('f, 'l) lab, 'v) rules"
                   "(('f, 'l) lab, 'v) rules"
                   "('f, 'l, 'v) reltrs_nontermination_proof"
| TRS_Confluence_Proof bool "(('f, 'l) lab, 'v) rules"
                 "('f, 'l, 'v) cr_proof"
| TRS_Non_Confluence_Proof bool "(('f, 'l) lab, 'v) rules"
                 "('f, 'l, 'v, 'v) ncr_proof"
| Completion_Proof "(('f, 'l) lab, 'v) equation list"
                   "(('f, 'l) lab, 'v) rules"
                   "('f, 'l, 'v) completion_proof"
| Equational_Proof "(('f, 'l) lab, 'v) equation list"
                   "(('f, 'l) lab, 'v) term \<times> (('f, 'l) lab, 'v) term"
                   "('f, 'l, 'v) equational_proof"
| Equational_Disproof "(('f, 'l) lab, 'v) equation list"
                   "(('f, 'l) lab, 'v) term \<times> (('f, 'l) lab, 'v) term"
                   "('f, 'l, 'v) equational_disproof"
| Quasi_Reductive_Proof "(('f, 'l) lab, 'v) crules"
                   "('f, 'l, 'v) quasi_reductive_proof"
| Unknown_Proof unknown_info "('f,'l,'v) unknown_proof"
| Unknown_Disproof unknown_info "('f,'l,'v) neg_unknown_proof"

hide_const (open) eq (* eq from HOL-Algebra congruence is here used as variable name for an equation *)

fun input :: "('f, 'l, 'v) cert_problem \<Rightarrow> (('f, 'l) lab, 'v) input_problem" where
  "input (TRS_Termination_Proof nfs Q R S_o _) = TRS nfs Q R (rel_rules_of S_o)"
| "input (Complexity_Proof Q R S_o cm cc _) = Complexity_TRS Q R (rel_rules_of S_o) cm cc"
| "input (TRS_Nontermination_Proof nfs Q R _) = TRS nfs Q R []"
| "input (Outermost_Nontermination_Proof R _) = OTRS R"
| "input (Outermost_Termination_Proof R _) = OTRS R"
| "input (CS_Nontermination_Proof \<mu> R _) = CSTRS \<mu> R"
| "input (CS_Termination_Proof \<mu> R _) = CSTRS \<mu> R"
| "input (FP_Termination_Proof P R _) = FPTRS P R"
| "input (FP_Nontermination_Proof P R _) = FPTRS P R"
| "input (Relative_TRS_Nontermination_Proof nfs Q R S _) = TRS nfs Q R S"
| "input (DP_Termination_Proof nfs m P Pw Q R Rw _) = DP nfs m P Pw Q R Rw"
| "input (DP_Nontermination_Proof nfs m P Q R _) = DP nfs m P [] Q [] R"
| "input (TRS_Confluence_Proof nfs R _) = TRS nfs No_Strategy R []"
| "input (TRS_Non_Confluence_Proof nfs R _) = TRS nfs No_Strategy R []"
| "input (Completion_Proof E R _) = EQS_TRS E R"
| "input (Equational_Proof E eq _) = EQS_EQ E eq"
| "input (Equational_Disproof E eq _) = EQS_EQ E eq"
| "input (Quasi_Reductive_Proof ctrs _) = CTRS ctrs"
| "input (Unknown_Proof up _) = Unknown up"
| "input (Unknown_Disproof up _) = Unknown up"

fun desired_property :: "('f, 'l, 'v) cert_problem \<Rightarrow> bool" where
  "desired_property (TRS_Termination_Proof nfs Q R S_o _ ) = SN_qrel (nfs, set (strategy_to_Q Q R), set R, set (rel_rules_of S_o))"
| "desired_property (Complexity_Proof Q R S_o cm cc _ ) = deriv_bound_measure_class (rel_qrstep (True, set (strategy_to_Q Q R), set R, set (rel_rules_of S_o))) cm cc"
| "desired_property (TRS_Nontermination_Proof nfs Q R _) = (\<not> SN (qrstep nfs (set (strategy_to_Q Q R)) (set R)))"
| "desired_property (Outermost_Nontermination_Proof R _) = (\<not> SN (ostep (lhss (set R)) (set R)))"
| "desired_property (Outermost_Termination_Proof R _) = (SN (ostep (lhss (set R)) (set R)))"
| "desired_property (FP_Nontermination_Proof P R _) = (\<not> SN (fpstep (set P) (set R)))"
| "desired_property (FP_Termination_Proof P R _) = (SN (fpstep (set P) (set R)))"
| "desired_property (CS_Nontermination_Proof \<mu> R _) = (\<not> SN (csstep (mu_of \<mu>) (set R)))"
| "desired_property (CS_Termination_Proof \<mu> R _) = (SN (csstep (mu_of \<mu>) (set R)))"
| "desired_property (Relative_TRS_Nontermination_Proof nfs Q R S _) = (\<not> SN_qrel (nfs, set (strategy_to_Q Q R), set R, set S))"
| "desired_property (DP_Termination_Proof nfs m P Pw Q R Rw _) = finite_dpp (nfs, m, set P, set Pw, set (strategy_to_Q Q (R @ Rw)), set R, set Rw)"
| "desired_property (DP_Nontermination_Proof nfs m P Q R _) = infinite_dpp (nfs, set P, set (strategy_to_Q Q R), set R)" (*ignores nfs and m flags!*)
| "desired_property (TRS_Confluence_Proof nfs R _) = CR (rstep (set R))"
| "desired_property (TRS_Non_Confluence_Proof nfs R _) = (\<not> CR (rstep (set R)))"
| "desired_property (Completion_Proof E R _) = completed_rewrite_system (set E) (set R)"
| "desired_property (Equational_Proof E eq _) = (eq \<in> (estep (set E))^*)"
| "desired_property (Equational_Disproof E eq _) = (eq \<notin> (estep (set E))^*)"
| "desired_property (Quasi_Reductive_Proof ctrs _) = (quasi_reductive (set ctrs))"
| "desired_property (Unknown_Proof up _) = (unknown_satisfied up)"
| "desired_property (Unknown_Disproof up _) = (unknown_satisfied up)"

fun xml_tag :: "('f, 'l, 'v) cert_problem \<Rightarrow> string" where
  "xml_tag (TRS_Termination_Proof _ _ _ (Some _) _ ) = ''<relativeTerminationProof>''"
| "xml_tag (TRS_Termination_Proof _ _ _ None _ ) = ''<trsTerminationProof>''"
| "xml_tag (Complexity_Proof _ _ _ _ _ _ ) = ''<complexityProof>''"
| "xml_tag (TRS_Nontermination_Proof _ _ _ _) = ''<trsNonterminationProof>''"
| "xml_tag (Outermost_Nontermination_Proof _ _) = ''<trsNonterminationProof>''"
| "xml_tag (Outermost_Termination_Proof _ _) = ''<trsTerminationProof>''"
| "xml_tag (FP_Nontermination_Proof _ _ _) = ''<trsNonterminationProof>''"
| "xml_tag (FP_Termination_Proof _ _ _) = ''<trsTerminationProof>''"
| "xml_tag (CS_Nontermination_Proof _ _ _) = ''<trsNonterminationProof>''"
| "xml_tag (CS_Termination_Proof _ _ _) = ''<trsTerminationProof>''"
| "xml_tag (Relative_TRS_Nontermination_Proof _ _ _ _ _) = ''<relativeNonterminationProof>''"
| "xml_tag (DP_Termination_Proof _ _ _ _ _ _ _ _) = ''<dpProof>''"
| "xml_tag (DP_Nontermination_Proof _ _ _ _ _ _) = ''<dpNonterminationProof>''"
| "xml_tag (TRS_Confluence_Proof _ _ _) = ''<crProof>''"
| "xml_tag (TRS_Non_Confluence_Proof _ _ _) = ''<crDisproof>''"
| "xml_tag (Completion_Proof _ _ _) = ''<completionProof>''"
| "xml_tag (Equational_Proof _ _ _) = ''<equationalProof>''"
| "xml_tag (Equational_Disproof _ _ _) = ''<equationalDisproof>''"
| "xml_tag (Quasi_Reductive_Proof _ _) = ''<quasiReductiveProof>''"
| "xml_tag (Unknown_Proof _ _) = ''<unknownInputProof>''"
| "xml_tag (Unknown_Disproof _ _) = ''<unknownInputProof>''"

fun cert_assms :: "bool \<Rightarrow> ('f, 'l, 'v) cert_problem \<Rightarrow> (('f, 'l, 'v) assm) list" where
  "cert_assms a (TRS_Termination_Proof _ _ _ _ p) = sn_assms a p"
| "cert_assms a (Outermost_Termination_Proof _ p) = sn_fp_assms a p"
| "cert_assms a (Outermost_Nontermination_Proof _ p) = neg_fp_sn_assms a p"
| "cert_assms a (FP_Termination_Proof _ _ p) = sn_fp_assms a p"
| "cert_assms a (FP_Nontermination_Proof _ _ p) = neg_fp_sn_assms a p"
| "cert_assms a (CS_Termination_Proof _ _ p) = sn_fp_assms a p"
| "cert_assms a (CS_Nontermination_Proof _ _ p) = neg_fp_sn_assms a p"
| "cert_assms a (TRS_Confluence_Proof _ _ p) = cr_assms a p"
| "cert_assms a (TRS_Non_Confluence_Proof _ _ p) = ncr_assms a p"
| "cert_assms a (Completion_Proof _ _ p) = completion_assms a p"
| "cert_assms a (DP_Termination_Proof _ _ _ _ _ _ _ p) = dp_assms a p"
| "cert_assms a (Equational_Proof _ _ p) = equational_assms a p"
| "cert_assms a (Equational_Disproof _ _ p) = equational_dis_assms a p"
| "cert_assms a (Quasi_Reductive_Proof _ p) = quasi_reductive_assms a p"
| "cert_assms a (Unknown_Proof _ p) = unknown_assms a p"
| "cert_assms a (Unknown_Disproof _ p) = neg_unknown_assms a p"
| "cert_assms a (Relative_TRS_Nontermination_Proof _ _ _ _ p) = neg_rel_sn_assms a p"
| "cert_assms a (TRS_Nontermination_Proof _ _ _ p) = neg_sn_assms a p"
| "cert_assms a (DP_Nontermination_Proof _ _ _ _ _ p) = neg_dp_assms a p"
| "cert_assms _ _ = []"

lemma cert_assms_False[simp]: "cert_assms False p = []" by (cases p) simp_all

datatype cert_result = Certified string | Unsupported string | Error string

fun
  certify_cert_problem ::
    "bool \<Rightarrow> ('tp, ('f::{show,key,countable}, label_type) lab, string) tp_ops \<Rightarrow>
     ('dpp, ('f, label_type) lab, string) dpp_ops \<Rightarrow> 
     ('f, label_type, string) cert_problem \<Rightarrow> cert_result"
where
  "certify_cert_problem a I J (TRS_Termination_Proof nfs Q R None prf) =
    (case check_trs_termination_proof I J a (shows_string ''1'') (tp_ops.mk I nfs (strategy_to_Q Q R) R []) prf of
      Inl err \<Rightarrow> Error ((''termination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (TRS_Termination_Proof nfs Q R None prf)))"
| "certify_cert_problem a I J (TRS_Termination_Proof nfs Q R (Some S) prf) = 
    (case check_trs_termination_proof I J a (shows_string ''1'') (tp_ops.mk I nfs (strategy_to_Q Q R) R S) prf of
      Inl err \<Rightarrow> Error ((''relative termination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (TRS_Termination_Proof nfs Q R (Some S) prf)))"
| "certify_cert_problem a I J (Complexity_Proof Q R S_o cm cc prf) = 
    (case check_complexity_proof (shows_string ''1'') I (tp_ops.mk I True (strategy_to_Q Q R) R (rel_rules_of S_o)) cm cc prf of
      Inl err \<Rightarrow> Error ((''complexity proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Complexity_Proof Q R S_o cm cc prf)))"
| "certify_cert_problem a I J (TRS_Nontermination_Proof nfs Q R prf) =
    (case check_trs_nontermination_proof I J a (shows_string ''1'') (tp_ops.mk I nfs (strategy_to_Q Q R) R []) prf of
      Inl err \<Rightarrow> Error ((''nontermination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (TRS_Nontermination_Proof nfs Q R prf)))"
| "certify_cert_problem a I J (Outermost_Nontermination_Proof R prf) =
    (case check_fp_nontermination_proof I J a (shows_string ''1'') (o_to_fp_impl (map fst R),R) prf of
      Inl err \<Rightarrow> Error ((''outermost nontermination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Outermost_Nontermination_Proof R prf)))"
| "certify_cert_problem a I J (Outermost_Termination_Proof R prf) =
    (case check_fptrs_termination_proof I J a (shows_string ''1'') (o_to_fp_impl (map fst R), R) prf of
      Inl err \<Rightarrow> Error ((''outermost termination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Outermost_Termination_Proof R prf)))"
| "certify_cert_problem a I J (CS_Nontermination_Proof \<mu> R prf) =
    (case check_fp_nontermination_proof  I J a (shows_string ''1'') (mu_to_fp_impl \<mu>,R) prf of
      Inl err \<Rightarrow> Error ((''outermost nontermination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (CS_Nontermination_Proof \<mu> R prf)))"
| "certify_cert_problem a I J (CS_Termination_Proof \<mu> R prf) =
    (case check_fptrs_termination_proof I J a (shows_string ''1'') (mu_to_fp_impl \<mu>, R) prf of
      Inl err \<Rightarrow> Error ((''outermost termination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (CS_Termination_Proof \<mu> R prf)))"
| "certify_cert_problem a I J (FP_Termination_Proof P R prf) =
    (case check_fptrs_termination_proof I J a (shows_string ''1'') (P, R) prf of
      Inl err \<Rightarrow> Error ((''forbidden pattern termination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (FP_Termination_Proof P R prf)))"
| "certify_cert_problem a I J (FP_Nontermination_Proof P R prf) =
    (case check_fp_nontermination_proof  I J a (shows_string ''1'') (P,R) prf of
      Inl err \<Rightarrow> Error ((''forbidden pattern nontermination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (FP_Nontermination_Proof P R prf)))"
| "certify_cert_problem a I J (Relative_TRS_Nontermination_Proof nfs Q R S prf) =
    (case check_reltrs_nontermination_proof I J a (shows_string ''1'') (tp_ops.mk I nfs (strategy_to_Q Q R) R S) prf of
      Inl err \<Rightarrow>
        Error ((''relative nontermination proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Relative_TRS_Nontermination_Proof nfs Q R S prf)))"
| "certify_cert_problem a I J (DP_Termination_Proof nfs m P Pw Q R Rw prf) =
    (case check_dp_termination_proof I J a (shows_string ''1'') (dpp_ops.mk J nfs m P Pw (strategy_to_Q Q (R @ Rw)) R Rw) prf of
      Inl err \<Rightarrow> Error ((''finiteness proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (DP_Termination_Proof nfs m P Pw Q R Rw prf)))"
| "certify_cert_problem a I J (DP_Nontermination_Proof nfs m P Q R prf) =
    (case check_dp_nontermination_proof I J a (shows_string ''1'') (dpp_ops.mk J nfs m P [] (strategy_to_Q Q R) [] R) prf of
      Inl err \<Rightarrow> Error ((''infiniteness proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (DP_Nontermination_Proof nfs m P Q R prf)))"
| "certify_cert_problem a I J (TRS_Confluence_Proof nfs R prf) =
    (case check_cr_proof a (shows_string ''1'') I J R prf of
      Inl err \<Rightarrow> Error ((''confluence proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (TRS_Confluence_Proof nfs R prf)))"
| "certify_cert_problem a I J (TRS_Non_Confluence_Proof nfs R prf) =
    (case check_ncr_proof a (shows_string ''1'') I J R prf of
      Inl err \<Rightarrow> Error ((''confluence disproof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (TRS_Non_Confluence_Proof nfs R prf)))"
| "certify_cert_problem a I J (Completion_Proof E R prf) =
    (case check_completion_proof a (shows_string ''1'') I J E R prf of
      Inl err \<Rightarrow> Error ((''completion proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Completion_Proof E R prf)))"
| "certify_cert_problem a I J (Equational_Proof E eq prf) =
    (case check_equational_proof a (shows_string ''1'') I J E eq prf of
      Inl err \<Rightarrow> Error ((''equational proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Equational_Proof E eq prf)))"
| "certify_cert_problem a I J (Equational_Disproof E eq prf) =
    (case check_equational_disproof a (shows_string ''1'') I J E eq prf of
      Inl err \<Rightarrow> Error ((''equational disproof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Equational_Disproof E eq prf)))"
| "certify_cert_problem a I J (Quasi_Reductive_Proof ctrs prf) =
    (case check_quasi_reductive_proof a (shows_string ''1'') I J ctrs prf of
      Inl err \<Rightarrow> Error ((''quasi reductive proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Quasi_Reductive_Proof ctrs prf)))"
| "certify_cert_problem a I J (Unknown_Proof up prf) =
    (case check_unknown_proof I J a (shows_string ''1'') up prf of
      Inl err \<Rightarrow> Error ((''unknown proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Unknown_Proof up prf)))"
| "certify_cert_problem a I J (Unknown_Disproof up prf) =
    (case check_unknown_disproof I J a (shows_string ''1'') up prf of
      Inl err \<Rightarrow> Error ((''unknown proof not accepted'' +#+ shows_nl +@+ err) [])
    | Inr _ \<Rightarrow> Certified (xml_tag (Unknown_Disproof up prf)))"


lemma certify_cert_problem_with_assms_sound:
  assumes I: "tp_spec I" and J: "dpp_spec J"
    and answer: "certify_cert_problem a I J cp = Certified info"
  shows "((\<forall>p\<in>set (cert_assms a cp). holds p) \<longrightarrow> desired_property cp) \<and> xml_tag cp = info"
proof -
  let ?i = "shows_string ''1''"
  interpret tp_spec I by (rule I)  
  show ?thesis
  proof (cases cp)
    case (TRS_Termination_Proof nfs Q R sOpt prof)
    let ?tp = "tp_ops.mk I nfs (strategy_to_Q Q R) R"
    show ?thesis
    proof (cases sOpt)
      case None
      from answer have ok: "isOK (check_trs_termination_proof I J a ?i (?tp []) prof) \<and> xml_tag cp = info" 
        unfolding TRS_Termination_Proof None
        by (cases "check_trs_termination_proof I J a ?i (?tp []) prof") auto
      from check_trs_termination_proof_with_assms[OF I J ok[THEN conjunct1]]
       and ok show ?thesis unfolding TRS_Termination_Proof None by simp
    next
      case (Some S)
      from answer have ok: "isOK (check_trs_termination_proof I J a ?i (?tp S) prof) \<and> xml_tag cp = info"
        unfolding TRS_Termination_Proof Some
        by (cases "check_trs_termination_proof I J a ?i (?tp S) prof", auto)
      from check_trs_termination_proof_with_assms[OF I J ok[THEN conjunct1]] 
        and ok show ?thesis unfolding TRS_Termination_Proof Some by simp
    qed
  next
    case (Complexity_Proof Q R S_o cm cc prof)
    let ?Q = "strategy_to_Q Q R"
    let ?S = "rel_rules_of S_o"
    let ?tp = "tp_ops.mk I True ?Q R ?S"
    from answer have ok: "isOK (check_complexity_proof ?i I ?tp cm cc prof) \<and> xml_tag cp = info" 
      unfolding Complexity_Proof 
      by (cases "check_complexity_proof ?i I ?tp cm cc prof") auto
    from check_complexity_proof_sound[OF I ok[THEN conjunct1]]
      and ok show ?thesis unfolding Complexity_Proof by simp
  next
    case (TRS_Nontermination_Proof nfs Q R "prf")
    let ?tp = "tp_ops.mk I nfs (strategy_to_Q Q R) R []"
    from answer have ok: "isOK (check_trs_nontermination_proof I J a ?i ?tp prf) \<and> xml_tag cp = info"
      unfolding TRS_Nontermination_Proof
      by (cases "check_trs_nontermination_proof I J a ?i ?tp prf") auto
    from check_trs_nontermination_proof_with_assms[OF I J _ ok[THEN conjunct1]]
      ok show ?thesis unfolding mk_simps TRS_Nontermination_Proof by simp
  next
    case (Outermost_Termination_Proof R "prf")
    have id: "lhss (set R) = set (map fst R)" by auto
    let ?tp = "(o_to_fp_impl (map fst R),R)"
    from answer have ok: "isOK (check_fptrs_termination_proof I J a ?i ?tp prf) \<and> xml_tag cp = info"
      unfolding Outermost_Termination_Proof
      by (cases "check_fptrs_termination_proof I J a ?i ?tp prf") auto
    from check_fptrs_termination_proof_with_assms[OF I J ok[THEN conjunct1]]
      ok show ?thesis unfolding Outermost_Termination_Proof 
      by (simp add: SN_fpstep_def, unfold id ostep_iff_fpstep_impl, simp)
  next
    case (Outermost_Nontermination_Proof R "prf")
    have id: "lhss (set R) = set (map fst R)" by auto
    let ?p = "o_to_fp_impl (map fst R)"
    from answer have ok: "isOK (check_fp_nontermination_proof I J a ?i (?p,R) prf) \<and> xml_tag cp = info"
      unfolding Outermost_Nontermination_Proof
      by (cases "check_fp_nontermination_proof I J a ?i (?p,R) prf") auto
    from check_fp_nontermination_proof_with_assms[OF I J _ ok[THEN conjunct1]]
      ok show ?thesis unfolding mk_simps Outermost_Nontermination_Proof 
      by (simp add: SN_fpstep_def, unfold id ostep_iff_fpstep_impl, simp)
  next
    case (CS_Termination_Proof \<mu> R "prf")
    let ?tp = "(mu_to_fp_impl \<mu>,R)"
    from answer have ok: "isOK (check_fptrs_termination_proof I J a ?i ?tp prf) \<and> xml_tag cp = info"
      unfolding CS_Termination_Proof
      by (cases "check_fptrs_termination_proof I J a ?i ?tp prf") auto
    from check_fptrs_termination_proof_with_assms[OF I J ok[THEN conjunct1]]
      ok show ?thesis unfolding CS_Termination_Proof by (simp add: SN_fpstep_def csstep_iff_fpstep_impl)
  next
    case (CS_Nontermination_Proof \<mu> R "prf")
    let ?p = "mu_to_fp_impl \<mu>"
    from answer have ok: "isOK (check_fp_nontermination_proof I J a ?i (?p,R) prf) \<and> xml_tag cp = info"
      unfolding CS_Nontermination_Proof
      by (cases "check_fp_nontermination_proof I J a ?i (?p,R) prf") auto
    from check_fp_nontermination_proof_with_assms[OF I J _ ok[THEN conjunct1]]
      ok show ?thesis unfolding mk_simps CS_Nontermination_Proof by (simp add: SN_fpstep_def csstep_iff_fpstep_impl)
  next
    case (FP_Termination_Proof P R "prf")
    let ?tp = "(P,R)"
    from answer have ok: "isOK (check_fptrs_termination_proof I J a ?i ?tp prf) \<and> xml_tag cp = info"
      unfolding FP_Termination_Proof
      by (cases "check_fptrs_termination_proof I J a ?i ?tp prf") auto
    from check_fptrs_termination_proof_with_assms[OF I J ok[THEN conjunct1]]
      ok show ?thesis unfolding FP_Termination_Proof by (simp add: SN_fpstep_def)
  next
    case (FP_Nontermination_Proof P R "prf")
    from answer have ok: "isOK (check_fp_nontermination_proof I J a ?i (P,R) prf) \<and> xml_tag cp = info"
      unfolding FP_Nontermination_Proof
      by (cases "check_fp_nontermination_proof I J a ?i (P,R) prf") auto
    from check_fp_nontermination_proof_with_assms[OF I J _ ok[THEN conjunct1]]
      ok show ?thesis unfolding mk_simps FP_Nontermination_Proof by (simp add: SN_fpstep_def)
  next
    case (Relative_TRS_Nontermination_Proof nfs Q R S "prf")
    let ?tp = "tp_ops.mk I nfs (strategy_to_Q Q R) R S"
    from answer have ok: "isOK (check_reltrs_nontermination_proof I J a ?i ?tp prf) \<and> xml_tag cp = info"
      unfolding Relative_TRS_Nontermination_Proof
      by (cases "check_reltrs_nontermination_proof I J a ?i ?tp prf") auto
    from check_reltrs_nontermination_proof_with_assms[OF I J _ ok[THEN conjunct1]]
      ok show ?thesis unfolding mk_simps Relative_TRS_Nontermination_Proof by simp
  next
    case (DP_Termination_Proof nfs m P Pw Q R Rw "prf")
    let ?dp = "dpp_ops.mk J nfs m P Pw (strategy_to_Q Q (R @ Rw)) R Rw"
    from answer have ok: "isOK (check_dp_termination_proof I J a ?i ?dp prf) \<and> xml_tag cp = info"
      unfolding DP_Termination_Proof
      by (cases "check_dp_termination_proof I J a ?i ?dp prf", auto)
    interpret J: dpp_spec J by (rule J)
    show ?thesis
    proof (intro conjI impI)
      assume ass: "\<forall> p \<in> set (cert_assms a cp). holds p"
      show "desired_property cp"
        unfolding DP_Termination_Proof desired_property.simps
      proof (rule check_dp_termination_proof_with_assms[OF I J ok[THEN conjunct1], unfolded 
            J.dpp_spec_sound J.mk_simps cert_assms.simps])
        from ass
        show "\<forall>p \<in> set (dp_assms a prf). holds p"
          unfolding DP_Termination_Proof by simp
      qed
    qed (insert ok, auto)
  next
    case (DP_Nontermination_Proof nfs m P Q R "prf")
    let ?dp = "dpp_ops.mk J nfs m P [] (strategy_to_Q Q R) [] R"
    from answer have ok: "isOK (check_dp_nontermination_proof I J a ?i ?dp prf) \<and> xml_tag cp = info"
      unfolding DP_Nontermination_Proof
      by (cases "check_dp_nontermination_proof I J a ?i ?dp prf", auto)
    interpret J: dpp_spec J by (rule J)
    show ?thesis
    proof (intro conjI impI)
      assume ass: "\<forall> p \<in> set (cert_assms a cp). holds p"
      show "desired_property cp"
        unfolding DP_Nontermination_Proof desired_property.simps
        using check_dp_nontermination_proof_with_assms[OF I J _ ok[THEN conjunct1], unfolded 
            J.dpp_spec_sound J.mk_simps]
        using DP_Nontermination_Proof ass
        by auto
    qed (insert ok, auto)
  next
    case (TRS_Confluence_Proof nfs R "prf")
    from answer have ok: "isOK (check_cr_proof a ?i I J R prf) \<and> xml_tag cp = info"
      unfolding TRS_Confluence_Proof
      by (cases "check_cr_proof a ?i I J R prf", auto)
    note res = check_cr_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding TRS_Confluence_Proof by auto
  next
    case (TRS_Non_Confluence_Proof nfs R "prf")
    from answer have ok: "isOK (check_ncr_proof a ?i I J R prf) \<and> xml_tag cp = info"
      unfolding TRS_Non_Confluence_Proof
      by (cases "check_ncr_proof a ?i I J R prf", auto)
    note res = check_ncr_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding TRS_Non_Confluence_Proof by auto
  next
    case (Completion_Proof E R "prf")
    from answer have ok: "isOK (check_completion_proof a ?i I J E R prf) \<and> xml_tag cp = info"
      unfolding Completion_Proof
      by (cases "check_completion_proof a ?i I J E R prf", auto)
    note res = check_completion_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding Completion_Proof by auto
  next
    case (Equational_Proof E eq "prf")
    from answer have ok: "isOK (check_equational_proof a ?i I J E eq prf) \<and> xml_tag cp = info"
      unfolding Equational_Proof by (cases "check_equational_proof a ?i I J E eq prf") auto
    note res = check_equational_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding Equational_Proof by auto
  next
    case (Equational_Disproof E eq "prf")
    from answer have ok: "isOK (check_equational_disproof a ?i I J E eq prf) \<and> xml_tag cp = info"
      unfolding Equational_Disproof by (cases "check_equational_disproof a ?i I J E eq prf") auto
    note res = check_equational_disproof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding Equational_Disproof by auto
  next
    case (Quasi_Reductive_Proof ctrs "prf")
    from answer have ok: "isOK (check_quasi_reductive_proof a ?i I J ctrs prf) \<and> xml_tag cp = info"
      unfolding Quasi_Reductive_Proof by (cases "check_quasi_reductive_proof a ?i I J ctrs prf") auto
    note res = check_quasi_reductive_proof_with_assms_sound[OF I J _ ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding Quasi_Reductive_Proof by auto
  next
    case (Unknown_Proof up "prf")
    from answer have ok: "isOK (check_unknown_proof I J a ?i up prf) \<and> xml_tag cp = info"
      unfolding Unknown_Proof by (cases "check_unknown_proof I J a ?i up prf") auto
    note res = check_unknown_proof_with_assms[OF I J ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding Unknown_Proof by auto
  next
    case (Unknown_Disproof up "prf")
    from answer have ok: "isOK (check_unknown_disproof I J a ?i up prf) \<and> xml_tag cp = info"
      unfolding Unknown_Disproof by (cases "check_unknown_disproof I J a ?i up prf") auto
    note res = check_unknown_disproof_with_assms[OF I J ok[THEN conjunct1]]
    show ?thesis using ok[THEN conjunct2] res unfolding Unknown_Disproof by auto
  qed
qed

lemma certify_cert_problem_sound:
  assumes I: "tp_spec I" and J: "dpp_spec J"
    and answer: "certify_cert_problem False I J cp = Certified info"
  shows "desired_property cp \<and> xml_tag cp = info"
  using certify_cert_problem_with_assms_sound[OF I J answer] by simp

end

