(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory RPO
imports
  Term_Order_Impl
  Lexicographic_Extension
  Term_Impl
  Reduction_Pair
  Multiset_Extension2
begin


(* precedence, max arity, tagging function, and two terms are given,
 result is pair (strict, non-strict) *)

datatype order_tag = Lex | Mul

function rpo  :: "('f \<Rightarrow> nat \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> bool \<times> bool) \<times> ('f \<Rightarrow> bool) \<Rightarrow> ('f \<Rightarrow> nat \<Rightarrow> order_tag) \<Rightarrow> nat \<Rightarrow>
                  ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool \<times> bool" 
where
  "rpo _ _ _ (Var x) (Var y) = (False, x = y)" |
  "rpo pr _ _ (Var x) (Fun g ts) = (False, ts = [] \<and> (snd pr) g)" |
  "rpo pr c n (Fun f ss) (Var y) = (let con = (\<exists> s \<in> set ss. snd (rpo pr c n s (Var y))) in (con,con))" |
  "rpo pr c n (Fun f ss) (Fun g ts) = (
    if (\<exists> s \<in> set ss. snd (rpo pr c n s (Fun g ts)))
       then (True,True)
       else (let (prs,prns) = (fst pr) f (length ss) g (length ts) in 
         if prns \<and> (\<forall> t \<in> set ts. fst (rpo pr c n (Fun f ss) t))
         then if prs
              then (True,True) 
              else if c f (length ss) = Lex \<and> c g (length ts) = Lex
                   then lex_ext (rpo pr c n) n ss ts
                   else if c f (length ss) = Mul \<and> c g (length ts) = Mul
                        then mul_ext (rpo pr c n) ss ts
                        else (length ss \<noteq> 0 \<and> length ts = 0, length ts = 0)
          else (False,False)))"
by pat_completeness auto

termination rpo 
proof (relation "measures [\<lambda> (pr,c,n,s,t). size s + size t]")
{
  fix "pr" c n and f :: 'f and ss :: "('f,'v)term list" and g :: 'f and ts :: "('f,'v)term list" and x xa y xb x'
  assume s: "xb \<in> set ss" and t: "x' \<in> set ts"
  from s have s1: "size xb \<le> list_size size ss" by (induct ss, auto)
  from t have t1: "size x' \<le> list_size size ts" by (induct ts, auto)
  from s1 t1 s t show "((pr, c, n, xb, x'), pr, c, n, Fun f ss, Fun g ts) \<in> measures [\<lambda> (pr,c,n,s,t). size s + size t]"
    by auto
}
moreover
{
  fix "pr" c n and f :: 'f and ss :: "('f,'v)term list" and g :: 'f and ts :: "('f,'v)term list" and x xa xb xc i
  assume "i < length ss" and "i < length ts"
  then obtain s t where si: "s = ss ! i" and s: "s \<in> set ss" and ti: "t = ts ! i" and t: "t \<in> set ts" by auto
  from s have s1: "size s \<le> list_size size ss" by (induct ss, auto)
  from t have t1: "size t \<le> list_size size ts" by (induct ts, auto)
  from s1 t1 si ti show "((pr, c, n, ss ! i, ts ! i), pr, c, n, Fun f ss, Fun g ts) \<in> measures [\<lambda> (pr,c,n,s,t). size s + size t]"
    by auto
}
qed (auto simp: size_simps)

(* The strict order implies the non strict one *)
lemma rpo_stri_imp_nstri[rule_format]: "fst (rpo pr c n s t) \<longrightarrow>
  snd (rpo pr c n s t)"
proof (induct rule: rpo.induct[of "\<lambda> pr c n s t. fst (rpo pr c n s t) \<longrightarrow> snd (rpo pr c n s t)"])
  case (4 "pr" c n f ss g ts)
  obtain prc prl where prec: "pr = (prc,prl)"  by (cases "pr", auto)
  obtain  s ns where  prc: "prc f (length ss) g (length ts) = (s,ns)"
   by (cases "prc f (length ss) g (length ts)", auto)
  show ?case
    by (cases "length ss = length ts \<or> length ts \<le> n", auto simp: Let_def lex_ext_stri_imp_nstri
      mul_ext_stri_imp_nstri prec prc)
qed (auto simp: Let_def)

locale rpo_pr =
  fixes prc :: "'f \<Rightarrow> nat \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> bool \<times> bool"
  and   prl :: "'f \<Rightarrow> bool"
  and   c :: "'f \<Rightarrow> nat \<Rightarrow> order_tag"
  assumes prc_refl: "prc f n f n = (False,True)"
  and     prc_SN: "SN {((f,n),(g,m)) | f n g m. fst (prc f n g m)}"
  and     prl: "prl g \<Longrightarrow> snd (prc f n g 0) = True"
  and     prl3: "prl f \<Longrightarrow> snd (prc f 0 g 0) \<Longrightarrow> prl g"
  and  prc_compat: "prc f n g m = (s1,ns1) \<Longrightarrow> prc g m h k = (s2,ns2) \<Longrightarrow> prc f n h k = (s,ns) \<Longrightarrow> (ns1 \<and> ns2 \<longrightarrow> ns) \<and> (ns1 \<and> s2 \<longrightarrow> s) \<and> (s1 \<and> ns2 \<longrightarrow> s)"
  and  prc_stri_imp_nstri: "fst (prc f n g m) \<Longrightarrow> snd (prc f n g m)"
begin

lemma prl2: assumes g: "prl g" shows "fst (prc g 0 f n) = False"
proof(rule ccontr)
  assume "\<not> ?thesis"
  then obtain b where gf: "prc g 0 f n = (True,b)" by (cases "prc g 0 f n", auto)
  obtain b1 b2 where gg: "prc g 0 g 0 = (b1,b2)" by (cases "prc g 0 g 0", auto)
  obtain b' where fg: "prc f n g 0 = (b',True)" using prl[OF g, of f n] by (cases "prc f n g 0", auto)
  from prc_compat[OF gf fg gg] gg have gg: "fst (prc g 0 g 0)" by auto
  with SN_on_irrefl[OF prc_SN] show False by blast
qed

abbreviation "pr" where "pr \<equiv> (prc,prl)"
abbreviation rpo_pr where "rpo_pr \<equiv> rpo pr c"

(* Reflexivity of the non strict order *)

lemma rpo_nstri_refl: 
  shows "snd (rpo_pr n s s)"
proof (induct s)
let ?ns ="{(x, y). snd (rpo_pr n x y)}"
let ?s = "{(x, y). fst (rpo_pr n x y)}"
  case (Fun f ss)
  have rec11: "snd (lex_ext (rpo_pr n) n ss ss)"
    by (rule all_nstri_imp_lex_nstri, auto simp: Fun)
  have rec12: "snd (mul_ext (rpo_pr n) ss ss)"
   using ns_mul_ext_refl[of ?ns "multiset_of ss" ?s] and Fun and in_multiset_in_set[of ss]
   unfolding mul_ext_def Let_def snd_conv locally_refl_def by simp
  have rec21: "\<forall> s \<in> set ss. fst (rpo_pr n (Fun f ss) s)"
  proof
    fix s
    assume s: "s \<in> set ss"
    with Fun have rec: "snd (rpo_pr n s s)" by auto
    with s show "fst (rpo_pr n (Fun f ss) s)"
    proof (cases s)
      case (Fun g ts)
      from s rec have "\<exists> t \<in> set ss. snd (rpo_pr n t s)" by auto
      with Fun show ?thesis by auto
    next
      case (Var x)
      from s rec have "\<exists> t \<in> set ss. snd (rpo_pr n t s)" by auto
      with Var show ?thesis by auto
    qed
  qed
  have rec22: "\<forall> s \<in> set ss. fst (rpo_pr n (Fun f ss) s)"
  proof
    fix s
    assume s: "s \<in> set ss"
    with Fun have rec: "snd (rpo_pr n s s)" by auto
    with s show "fst (rpo_pr n (Fun f ss) s)"
    proof (cases s)
      case (Fun g ts)
      from s rec have "\<exists> t \<in> set ss. snd (rpo_pr n t s)" by auto
      with Fun show ?thesis by auto
    next
      case (Var x)
      from s rec have "\<exists> t \<in> set ss. snd (rpo_pr n t s)" by auto
      with Var show ?thesis by auto
    qed
  qed
  from rec11 rec12 rec21 rec22 show ?case by (cases "c f (length ss)", auto simp: Let_def prc_refl)
qed simp

(* Compatibility of the subterm relation with the order relation:
    a strict subterm is stricly smaller *)
lemma supt_imp_rpo_stri: assumes "s \<rhd> t" 
  shows "fst (rpo_pr n s t)"
  using assms
proof (induct)
  case (arg s ss f)
  have refl: "snd (rpo_pr n s s)" by (rule rpo_nstri_refl)
  from arg refl have fact: "\<exists> t \<in> set ss. snd (rpo_pr n t s)" by auto
  thus ?case 
    by (cases s, auto)
next
  case (subt s ss t f)
  from subt rpo_stri_imp_nstri[of "pr" c n s t] have "\<exists> s \<in> set ss. snd (rpo_pr n s t)" by auto
  thus ?case
    by (cases t, auto)
qed

(* Compatibility of the subterm relation with the order relation:
    a subterm is smaller *)
lemma supteq_imp_rpo_nstri: assumes supteq: "s \<unrhd> t" shows "snd (rpo_pr n s t)" 
proof -
  from supteq have "s = t \<or> s \<rhd> t" by auto
  thus ?thesis
  proof
    assume "s = t"
    with rpo_nstri_refl show ?thesis by auto
  next
    assume "s \<rhd> t"
    from rpo_stri_imp_nstri[OF supt_imp_rpo_stri[OF this]]
    show ?thesis .
  qed
qed

(* Monotonicity of the non-strict order with respect
    to (fun x \<rightarrow> Fun f (bef @ x # aft)) *)
lemma rpo_nstri_mono: assumes rel: "snd (rpo_pr n s t)"
  shows "snd (rpo_pr n (Fun f (bef @ s # aft)) (Fun f (bef @ t # aft)))"
proof -
  let ?ss = "bef @ s # aft"
  let ?ts = "bef @ t # aft"
  let ?s = "Fun f ?ss"
  let ?len = "Suc (length bef + length aft)"
  have aft_bef: "\<forall> b \<in> set bef \<union> set aft. fst (rpo_pr n ?s b)"
  proof (intro ballI)
    fix b
    assume elem: "b \<in> set bef \<union> set aft"
    show "fst (rpo_pr n ?s b)"
    proof (rule supt_imp_rpo_stri, cases "b \<in> set bef")
      assume "b \<in> set bef"
      thus "?s \<rhd> b" by (auto simp: supt.intros)
    next
      assume bbef: "b \<notin> set bef"
      show "?s \<rhd> b" 
      proof (cases "b \<in> set aft")
        assume "b \<notin> set aft"
        thus "?s \<rhd> b" using bbef elem by blast
      qed (auto simp: supt.intros)
    qed
  qed
  from rel have "\<exists> s \<in> set ?ss. snd (rpo_pr n s t)" by (auto)
  hence t: "fst (rpo_pr n ?s t)" by (cases t, auto)
  let ?map = "map (\<lambda> (s,t). rpo_pr n s t)"
  from length_append have sslts: "length ?ss = length ?ts" by simp
  have "(multiset_of ?ss, multiset_of ?ts) \<in> ns_mul_ext {(x, y). snd (rpo_pr n x y)}
       {(x, y). fst (rpo_pr n x y)}"
  proof (rule all_ns_ns_mul_ext)
  from sslts show "length ?ss = length ?ts" by assumption
  {
    fix i assume i_bd: "i < length ?ts"
    have "(?ss ! i, ?ts  ! i) \<in> {(x, y). snd (rpo_pr n x y)}"
    proof (cases "i < length bef")
    case True
      with append_Cons_nth_left[OF this] and rpo_nstri_refl
       show ?thesis by simp
    next
    case False note Hineq = this
      thus ?thesis proof (cases "i - length bef")
      case 0
        with False have "i = length bef" by simp
        with nth_append_length and rel show ?thesis by simp
      next
      case (Suc j)
        from nth_append and False
         have "\<And>u. (bef @ u # aft) ! i = (u # aft) ! (i - length bef)" by metis
        with Suc and rpo_nstri_refl show ?thesis by simp
      qed
    qed
  }
  thus "\<forall>i. i < length ?ts \<longrightarrow> (?ss ! i, ?ts ! i) \<in> {(x, y). snd (rpo_pr n x y)}" by simp
  qed
  hence mul_ext: "snd (mul_ext (rpo_pr n) ?ss ?ts)" unfolding mul_ext_def by simp
  have lex_ext: "snd (lex_ext (rpo_pr n) n ?ss ?ts)"
  proof (rule all_nstri_imp_lex_nstri[of ?ts _ ?ss, simplified], intro allI impI)
    fix i
    assume i: "i < ?len"
    show "snd (rpo_pr n (?ss ! i) (?ts ! i))" (is "?nstri i")
    proof (cases "i = length bef")
      case True
      with rel show "?nstri i" by auto
    next
      case False
      show ?thesis 
      proof (cases "i < length bef")
        case True
        with rpo_nstri_refl[of n] show "?nstri i" by (simp add: nth_append)
      next
        case False
        with `i \<noteq> length bef` have "i - length bef > 0" by arith 
        then obtain j where i: "i - length bef = Suc j" by (cases "i - length bef", auto)
        with False rpo_nstri_refl[of n "aft ! j"] show "?nstri i" by (simp add: nth_append)
      qed
    qed
  qed
(** UGLY stuff. TODO: clean this? Make it more robust? *)
  show ?thesis by (simp only: rpo.simps Let_def, cases "\<exists> s \<in> set ?ss. snd (rpo_pr n s (Fun f ?ts))", simp, simp add: prc_refl t aft_bef mul_ext[simplified] lex_ext[simplified], cases "c f (length (bef @ s # aft))", simp_all)
qed

(* Monotonicity of the strict order with respect
    to (fun x \<rightarrow> Fun f (bef @ x # aft)) *)
lemma rpo_stri_mono: 
  assumes rel: "fst (rpo_pr n s t)"
  shows "fst (rpo_pr n (Fun f (bef @ s # aft)) (Fun f (bef @ t # aft)))"
proof -
  let ?ss = "bef @ s # aft"
  let ?ts = "bef @ t # aft"
  let ?s = "Fun f ?ss"
  have aft_bef: "\<forall> b \<in> set bef \<union> set aft. fst (rpo_pr n ?s b)"
  proof (intro ballI)
    fix b
    assume elem: "b \<in> set bef \<union> set aft"
    show "fst (rpo_pr n ?s b)"
      proof (rule supt_imp_rpo_stri, cases "b \<in> set bef")
        assume "b \<in> set bef" 
        thus "?s \<rhd> b" by (auto simp: supt.intros)
      next 
        assume bbef: "b \<notin> set bef"
        show "?s \<rhd> b" 
        proof (cases "b \<in> set aft")
	  assume "b \<notin> set aft"
	  thus "?s \<rhd> b" using bbef elem by blast
        qed (auto simp: supt.intros)
      qed
  qed
  from rpo_stri_imp_nstri[OF rel] have "\<exists> s \<in> set ?ss. snd (rpo_pr n s t)" by auto
  hence t: "fst (rpo_pr n ?s t)" by (cases t, auto)
  let ?map = "map (\<lambda> (s,t). rpo_pr n s t)"
  obtain stri nstri where st: "rpo_pr n s t = (stri,nstri)" by force
  with rel have stri: "stri = True" by simp
  let ?ba = "bef @ aft"
  have Hrew: "\<And>t. remove_nth (length bef) (bef @ t # aft) = ?ba" by simp
  have bef_bd: "\<And>t. length bef < length (bef @ t # aft)" by simp
  with rpo_nstri_refl
   have "\<forall>i. i < length ?ba \<longrightarrow> (?ba ! i, ?ba ! i) \<in> {(x, y). snd (rpo_pr n x y)}"
   by auto
  from all_ns_ns_mul_ext[OF _ this]
   have H: "(multiset_of ?ba, multiset_of ?ba) \<in> ns_mul_ext  {(x, y). snd (rpo_pr n x y)} {(x, y). fst (rpo_pr n x y)}"
   by metis
  from rel have s_s_t: "\<forall>b. b \<in># {#t#} \<longrightarrow> (\<exists>a. a \<in># {#s#} \<and> (a,b) \<in> {(x,y). fst (rpo_pr n x y)})" by simp
  from remove_nth_soundness[OF bef_bd[of s]] and Hrew and nth_append_length[of bef s aft]
   and insert_DiffM2[OF nth_mem_multiset_of[OF bef_bd[of s]]]
   have H1: "multiset_of ?ba + {#s#} = multiset_of ?ss" by auto
  from remove_nth_soundness[OF bef_bd[of t]] and Hrew and nth_append_length[of bef t aft]
   and insert_DiffM2[OF nth_mem_multiset_of[OF bef_bd[of t]]]
   have H2: "multiset_of ?ba + {#t#} = multiset_of ?ts" by auto
  have "{#s#} \<noteq> {#}" by simp
  from ns_s_mul_ext_union_multiset_l[OF H this s_s_t] and H1 and H2
   have "(multiset_of ?ss, multiset_of ?ts)
     \<in> s_mul_ext  {(x, y). snd (rpo_pr n x y)} {(x, y). fst (rpo_pr n x y)}" by simp
  hence mul_ext: "fst (mul_ext (rpo_pr n) ?ss ?ts)" unfolding mul_ext_def by simp
  {
    fix i
    assume i: "i < length bef"
    hence "snd (rpo_pr n (?ss ! i) (?ts ! i))" by (simp add: rpo_nstri_refl[of n] nth_append)
  }
  hence lex_ext: "fst (lex_ext (rpo_pr n) n ?ss ?ts)"
    by (simp add: lex_ext_iff, intro exI[of _ "length bef"], simp add: st stri)
  show ?thesis by (simp only: rpo.simps Let_def, cases "\<exists> s \<in> set ?ss. snd (rpo_pr n s (Fun f ?ts))", simp, simp add: prc_refl t aft_bef mul_ext lex_ext, cases "c f (length (bef @ s # aft))", simp_all)
qed

(* Stability of the orders *)
lemma rpo_stable: fixes \<sigma> :: "('f,'v)subst"
  shows "(fst (rpo_pr n s t) \<longrightarrow> fst (rpo_pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>))) \<and> (snd (rpo_pr n s t) \<longrightarrow> snd (rpo_pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)))"
  (is "?p s t")
proof -
  have "pr = pr \<longrightarrow> c = c \<longrightarrow> ?p s t"
  proof (induct rule: rpo.induct[of "\<lambda> p d n s t. p = pr \<longrightarrow> d = c \<longrightarrow> (fst (rpo_pr n s t) \<longrightarrow> fst (rpo_pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>))) \<and> (snd (rpo_pr n s t) \<longrightarrow> snd (rpo_pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)))"])
    case (1 _ _ n x y)
    show ?case by (auto simp: rpo_nstri_refl)
  next
    case (2 _ _ n x g ts)
    show ?case
    proof -
      show ?thesis 
      proof (cases "\<sigma> x")
        case (Var y)
        show ?thesis by (auto simp: Var)
      next
        case (Fun f ss)
        show ?thesis 
        proof (cases "prl g")
          case False thus ?thesis  by auto
        next
          case True
          obtain s ns where "prc f (length ss) g 0 = (s,ns)" by (cases "prc f (length ss) g 0", auto)             with prl[OF True, of f "length ss"] have prc: "prc f (length ss) g 0 = (s, True)" by auto
          from ns_mul_ext_bottom have "snd (mul_ext (rpo_pr n) ss [])" by (simp add: ns_mul_ext_def mul_ext_def)
          thus ?thesis by (auto simp: Fun prc all_nstri_imp_mul_nstri[of "[]", simplified]
           all_nstri_imp_lex_nstri[of "[]", simplified])
        qed
      qed
    qed
  next
    case (3 p d n f ss y)
    {
      assume "\<exists> s \<in> set ss. snd (rpo_pr n s (Var y))" and p: "p = pr" and d: "d = c"
      from this obtain s where s: "s \<in> set ss" and "snd (rpo_pr n s (Var y))" by force
      with 3 p d have "snd (rpo_pr n (s \<cdot> \<sigma>) (Var y \<cdot> \<sigma>))" by auto
      with s have "\<exists> s' \<in> set (map (\<lambda> s'. s' \<cdot> \<sigma>) ss). snd (rpo_pr n s' (Var y \<cdot> \<sigma>))" by auto
    }
    thus ?case by (cases "Var y \<cdot> \<sigma>", auto simp: Let_def)
  next
    case (4 p d n f ss g ts)
    let ?map = "map (\<lambda> s'. s' \<cdot> \<sigma>)"
    let ?t = "Fun g ts"
    let ?s = "Fun f ss"
    let ?ssig = "Fun f ss \<cdot> \<sigma>"
    let ?tsig = "Fun g ts \<cdot> \<sigma>"
    let ?ss = "?map ss"
    let ?ts = "?map ts"
    show ?case
    proof (cases "p = pr")
      case True note p = this
      show ?thesis
      proof (cases "d = c")
      case True note d = this
        show ?thesis
        proof (cases "\<exists> s \<in> set ss. snd (rpo_pr n s ?t)")
        case True
          from this obtain s where s: "s \<in> set ss" and "snd (rpo_pr n s ?t)" by auto
          with 4 p d have "snd (rpo_pr n (s \<cdot> \<sigma>) ?tsig)" by auto
          with s have "\<exists> s' \<in> set ?ss. snd (rpo_pr n s' ?tsig)" by auto
          thus ?thesis by auto
        next
        case False
          hence not1: "\<not> (\<exists> s \<in> set ss. snd (rpo_pr n s ?t))" by force
          show ?thesis
          proof (cases "snd (rpo_pr n ?s ?t)")
          case False
            with rpo_stri_imp_nstri have "\<not> (fst (rpo_pr n ?s ?t))" by blast
            with False show ?thesis by auto
          next
          case True note Hsnd = this
            with not1 have smallTs: "\<forall> t \<in> set ts. fst (rpo_pr n ?s t)" by (cases ?thesis, auto)
            obtain s ns where prc: "prc f (length ss) g (length ts) = (s,ns)"
             by (cases "prc f (length ss) g (length ts)", auto)
            from True smallTs have ge: "ns = True"
             by (cases ?thesis, auto simp: Let_def not1 prc)
            note prc = prc[unfolded ge]
            have "\<forall> t \<in> set ts. fst (rpo_pr n ?ssig (t \<cdot> \<sigma>))"
             by (intro ballI, rule 4(2)[THEN mp[OF _ p], THEN mp[OF _ d], THEN conjunct1, THEN mp[OF _ smallTs[THEN bspec]], unfolded p d, OF not1], auto simp: prc)
            hence smallTssig: "\<forall> t \<in> set ?ts. fst (rpo_pr n ?ssig t)" by auto
            show ?thesis
            proof (cases s)
            case True with smallTssig not1 show ?thesis by (auto simp: prc)
            next
            case False
              hence "s = False" by simp
              note prc = prc[unfolded this]
              show ?thesis
              proof (cases "c f (length ss) = c g (length ts)")
              case False note cfg = this
                thus ?thesis
                proof (cases "c f (length ss)")
                case Lex note cf = this
                  with cfg have cg: "c g (length ts) = Mul" by (cases "c g (length ts)") simp
                  {
                    assume Hfst: "fst (local.rpo_pr n (Fun f ss) (Fun g ts))"
                    from Hfst and smallTs have ss_ne: "map (\<lambda>t. t \<cdot> \<sigma>) ss \<noteq> []"
                     by (simp add: cf cg prc not1)
                    from Hfst and smallTs have ts_e: "map (\<lambda>t. t \<cdot> \<sigma>) ts = []"
                     by (simp add: cf cg prc not1)
                    from ss_ne and ts_e and prc and cf and cg
                     have "fst (local.rpo_pr n (Fun f ss \<cdot> \<sigma>) (Fun g ts \<cdot> \<sigma>))"
                     by simp
                  }
                  moreover
                  {
                    assume Hsnd: "snd (local.rpo_pr n (Fun f ss) (Fun g ts))"
                    from Hsnd and smallTs have ts_e: "map (\<lambda>t. t \<cdot> \<sigma>) ts = []"
                     by (simp add: cf cg prc not1)
                    with prc and cf and cg
                     have "snd (local.rpo_pr n (Fun f ss \<cdot> \<sigma>) (Fun g ts \<cdot> \<sigma>))"
                     by simp
                  }
                  ultimately show ?thesis by simp
                next
                case Mul note cf = this
                  with cfg have cg: "c g (length ts) = Lex" by (cases "c g (length ts)") simp_all
                  {
                    assume Hfst: "fst (local.rpo_pr n (Fun f ss) (Fun g ts))"
                    from Hfst and smallTs have ss_ne: "map (\<lambda>t. t \<cdot> \<sigma>) ss \<noteq> []"
                     by (simp add: cf cg prc not1)
                    from Hfst and smallTs have ts_e: "map (\<lambda>t. t \<cdot> \<sigma>) ts = []"
                     by (simp add: cf cg prc not1)
                    from ss_ne and ts_e and prc and cf and cg
                     have "fst (local.rpo_pr n (Fun f ss \<cdot> \<sigma>) (Fun g ts \<cdot> \<sigma>))"
                     by simp
                  }
                  moreover
                  {
                    assume Hsnd: "snd (local.rpo_pr n (Fun f ss) (Fun g ts))"
                    from Hsnd and smallTs have ts_e: "map (\<lambda>t. t \<cdot> \<sigma>) ts = []"
                     by (simp add: cf cg prc not1)
                    with prc and cf and cg
                     have "snd (local.rpo_pr n (Fun f ss \<cdot> \<sigma>) (Fun g ts \<cdot> \<sigma>))"
                     by simp
                  }
                  ultimately show ?thesis by simp
                qed
              next
              case True note cfg = this
                show ?thesis
                proof (cases "c f (length ss)")
                case Mul note cf = this
                  with cfg have cg: "c g (length ts) = Mul" by simp
                  from prc not1 smallTs Hsnd cf cg
                   have snd_mul: "snd (mul_ext (rpo_pr n) ss ts)" by auto
	          let ?ind = "\<lambda> s t. (fst (rpo_pr n s t) \<longrightarrow> fst (rpo_pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>))) \<and>
                   (snd (rpo_pr n s t) \<longrightarrow> snd (rpo_pr n (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)))"
                  have ind: "\<And> s t. s \<in> set ss \<Longrightarrow> t \<in> set ts \<Longrightarrow> ?ind s t"
                   by (rule 4(4)[unfolded p d, OF not1 refl, simplified], auto simp: prc smallTs cf cg)
                  have snd_mul2: "snd (mul_ext (rpo_pr n) ?ss ?ts)"
                   by (rule nstri_mul_ext_map, insert snd_mul ind, auto)
	          have nstri: "snd (rpo_pr n ?ssig ?tsig)"
                   by (simp add: Let_def prc smallTssig[simplified] snd_mul2 cf cg)
	          {
	            assume "fst (rpo_pr n ?s ?t)"
	            with not1 smallTs prc snd_mul cf cg
                     have fst_mul: "fst (mul_ext (rpo_pr n) ss ts)" by auto
                    have fst_mul2: "fst (mul_ext (rpo_pr n) ?ss ?ts)"
                     by (rule stri_mul_ext_map, insert fst_mul ind, auto)
	            have "fst (rpo_pr n ?ssig ?tsig)"
                     by (simp add: Let_def prc smallTssig[simplified] fst_mul2 cf cg)
	          } 
	          with nstri show ?thesis by simp
                next 
                case Lex note cf = this
                  with cfg have cg: "c g (length ts) = Lex" by simp
	          from prc not1 smallTs Hsnd cf cg
                   have snd_lex: "snd (lex_ext (rpo_pr n) n ss ts)" by auto
                  let ?ind = "\<lambda> i. (fst (rpo_pr n (ss ! i) (ts ! i)) \<longrightarrow> fst (rpo_pr n (ss ! i \<cdot> \<sigma>)
                   (ts ! i \<cdot> \<sigma>))) \<and> (snd (rpo_pr n (ss ! i) (ts ! i)) \<longrightarrow> snd (rpo_pr n (ss ! i \<cdot> \<sigma>)
                   (ts ! i \<cdot> \<sigma>)))"
                  have ind: "\<And> i. \<lbrakk>i < length ss; i < length ts\<rbrakk> \<Longrightarrow> ?ind i"
                  by (rule 4(3)[unfolded p d, OF not1 refl, simplified], auto simp: prc smallTs cf cg)
                  from snd_lex have eq_small: "length ss = length ts \<or> length ts \<le> n"
                   by (simp add: lex_ext_iff)
                  from ind snd_lex have snd_lex2: "snd (lex_ext (rpo_pr n) n ?ss ?ts)"
                   by (simp add: lex_ext_iff eq_small, auto)
	          have nstri: "snd (rpo_pr n ?ssig ?tsig)"
                   by (simp add: Let_def prc snd_lex smallTssig[simplified] snd_lex2 cf cg)
	          {
	            assume "fst (rpo_pr n ?s ?t)"
	            with not1 smallTs prc snd_lex cf cg have "fst (lex_ext (rpo_pr n) n ss ts)" by auto
                    with ind have fst_lex2: "fst (lex_ext (rpo_pr n) n ?ss ?ts)"
                     by (simp add: lex_ext_iff eq_small, auto)
                    have "fst (rpo_pr n ?ssig ?tsig)"
                     by (simp add: Let_def prc snd_lex smallTssig[simplified] fst_lex2 cf cg)
	          }
	          with nstri show ?thesis by simp
                qed
              qed
            qed
          qed
        qed
      next
      case False thus ?thesis by simp
      qed
    next
    case False thus ?thesis by simp
    qed
  qed
  thus ?thesis by simp
qed

(* Least element *)

lemma rpo_least_1: assumes "prl f" 
  shows "snd (rpo_pr n t (Fun f []))"
proof (cases t, simp add: assms)
  case (Fun g ts)
  obtain s ns where "prc g (length ts) f 0 = (s,ns)" by (cases "prc g (length ts) f 0", auto)
  with prl[OF assms, of g "length ts"] have prc: "prc g (length ts) f 0 = (s,True)" by auto
  show ?thesis
  proof (cases "c f 0")
  case Mul note cf = this
    show ?thesis
    proof (cases "c g (length ts)")
    case Mul note cg = this
      from ns_mul_ext_bottom have H: "snd (mul_ext (rpo_pr n) ts [])" unfolding mul_ext_def by (simp add: ns_mul_ext_def)
      thus ?thesis by (simp add: assms Fun prc cf cg)
    next
    case Lex note cg = this show ?thesis by (simp add: assms Fun prc cf cg)
    qed
  next
  case Lex note cf = this
    show ?thesis
    proof (cases "c g (length ts)")
    case Mul note cg = this show ?thesis by (simp add: assms Fun prc cf cg)
    next
    case Lex note cg = this show ?thesis unfolding Fun by (simp add: lex_ext_least_1 prc cf cg)
    qed
  qed
qed


lemma rpo_least_2: assumes "prl f"
  shows "\<not> fst (rpo_pr n (Fun f []) t)"
proof (cases t, simp)
  case (Fun g ts)
  obtain s ns where "prc f 0 g (length ts) = (s,ns)" by (cases "prc f 0 g (length ts)", auto)
  with prl2[OF assms, of g "length ts"] have prc: "prc f 0 g (length ts) = (False,ns)" by auto
  show ?thesis
  proof (cases "c f 0")
  case Mul note cf = this
    show ?thesis
    proof (cases "c g (length ts)")
    case Mul note cg = this
      from s_mul_ext_bottom_strict have "\<not> fst (mul_ext (rpo_pr n) [] ts)"
       unfolding mul_ext_def s_mul_ext_def by simp
      thus ?thesis unfolding Fun by (simp add: Let_def lex_ext_least_2 prc cf cg)
    next
    case Lex note cg = this show ?thesis by (simp add: Fun cf cg prc)
    qed
  next
  case Lex note cf = this
    show ?thesis
    proof (cases "c g (length ts)")
    case Mul note cg = this show ?thesis by (simp add: Fun cf cg prc)
    next
    case Lex note cg = this
      show ?thesis unfolding Fun by (simp add: Let_def lex_ext_least_2 prc cf cg)
    qed
  qed
qed


lemma rpo_least_3: assumes "prl f" 
  and "snd (rpo_pr n (Fun f []) t)"
  shows "snd (rpo_pr n u t)"
using assms
proof (cases t, simp)
  case (Fun g ts)
  obtain s ns where "prc f 0 g (length ts) = (s,ns)" by (cases "prc f 0 g (length ts)", auto)
  with prl2[OF `prl f`, of g "length ts"] have prc: "prc f 0 g (length ts) = (False,ns)" by auto
  show ?thesis
  proof (cases ts)
    case Nil
    with Fun assms prc have "prc f 0 g (length ts) = (False,True)" by (cases ?thesis, simp, simp add: Let_def)
    with prl3[OF `prl f`, of g] Nil have "prl g" by auto
    show ?thesis unfolding Fun Nil by (rule rpo_least_1[OF `prl g`])
  next
    case (Cons s tts)
    have "\<not> fst (rpo_pr n (Fun f []) s)" by (rule rpo_least_2[OF `prl f`])
    with `snd (rpo_pr n (Fun f []) t)` Fun Cons have False by (simp add: Let_def)
    thus ?thesis ..
  qed
qed


(* Transitivity / compatibility of the orders *)
lemma rpo_compat[rule_format]: 
  fixes s t u :: "('f,'v)term"
  shows "
  (snd (rpo_pr n s t) \<and> fst (rpo_pr n t u) \<longrightarrow> fst (rpo_pr n s u)) \<and>
  (fst (rpo_pr n s t) \<and> snd (rpo_pr n t u) \<longrightarrow> fst (rpo_pr n s u)) \<and>
  (snd (rpo_pr n s t) \<and> snd (rpo_pr n t u) \<longrightarrow> snd (rpo_pr n s u))" (is "?tran s t u")
proof (rule
    wf_induct[OF wf_measures[of "[\<lambda> (s,t,u). size s, \<lambda> (s,t,u). size t, \<lambda> (s,t,u). size u]"], 
    of "\<lambda> (s,t,u). ?tran s t u" "(s,t,u)", simplified], clarify)
  fix s t u :: "('f,'v)term"
  assume ind: "\<forall> s' t' u'. (size s' < size s \<longrightarrow> ?tran s' t' u') \<and> (size s' = size s \<and> (size t' < size t \<or> size t' = size t \<and> size u' < size u) \<longrightarrow> ?tran s' t' u')"
  let ?GR = "\<lambda> s t. fst (rpo_pr n s t)"
  let ?GE = "\<lambda> s t. snd (rpo_pr n s t)"
  show "?tran s t u"
  proof (cases t)
    case (Var x)
    { 
      assume "?GR t u"
      with Var have False by (cases u, auto)
    } 
    moreover
    {
      assume gr: "?GR s t" and ge: "?GE t u"
      have "?GR s u"
      proof (cases u)
	case (Var y)
	with ge `t = Var x` have "t = u" by auto
	with gr show ?thesis by auto
      next
	case (Fun h us)
	with ge Var have us: "us = []" and pri: "prl h" by auto
	from gr Var obtain f ss where s: "s = Fun f ss" by (cases s, auto)
	with gr Var obtain s' where s': "s' \<in> set ss" by (auto simp: Let_def)
	have "?GE s' u" by (simp only: Fun us, rule rpo_least_1, rule pri)
	with s' have "\<exists> s' \<in> set ss. ?GE s' u" by (auto)
	with s Fun show ?thesis by simp
      qed
    }
    moreover
    {
      assume ge1: "?GE s t" and ge2: "?GE t u"
      have "?GE s u"
      proof (cases u)
	case (Var y)
	with ge2 `t = Var x` have "t = u" by auto
	with ge1 show ?thesis by auto
      next
	case (Fun h us)
	with ge2 Var have us: "us = []" and pri: "prl h" by auto
	show ?thesis by (simp only: Fun us, rule rpo_least_1, rule pri) 
      qed
    }
    ultimately show ?thesis by blast
  next
    case (Fun g ts)
    let ?t = "Fun g ts"
    from Fun have t: "t = ?t" .
    show ?thesis 
    proof (cases s)
      case (Var x)
      {
	assume gr: "?GR s t"
	with Var Fun have False by auto
      }
      moreover
      {
	assume ge: "?GE s t" and gr: "?GR t u"
	with Var Fun have pri: "prl g" and "ts = []" by auto
	with gr Fun have False using rpo_least_2[of g n u] by auto
      }
      moreover
      {
	assume ge1: "?GE s t" and ge2: "?GE t u"
	with Var Fun have pri: "prl g" and empty: "ts = []" by auto
	have "?GE s u" by (rule rpo_least_3[of g n u], rule pri, simp add: empty[symmetric] Fun[symmetric], rule ge2)
      }
      ultimately show ?thesis by blast
    next
      case (Fun f ss)
      let ?s = "Fun f ss"
      from Fun have s: "s = ?s" .
      let ?s1 = "\<exists> s' \<in> set ss. ?GE s' t"
      let ?t1 = "\<exists> t' \<in> set ts. ?GE t' u"
      let ?ls = "length ss"
      let ?lt = "length ts"
      obtain ps pns where prc: "prc f ?ls g ?lt = (ps,pns)" by (cases "prc f ?ls g ?lt")
      let ?tran2 = "\<lambda> a b c.  
         (snd (rpo_pr n a b) \<and> fst (rpo_pr n b c) \<longrightarrow> fst (rpo_pr n a c)) \<and>
         (fst (rpo_pr n a b) \<and> snd (rpo_pr n b c) \<longrightarrow> fst (rpo_pr n a c)) \<and> 
         (snd (rpo_pr n a b) \<and> snd (rpo_pr n b c) \<longrightarrow> snd (rpo_pr n a c)) \<and> 
         (fst (rpo_pr n a b) \<and> fst (rpo_pr n b c) \<longrightarrow> fst (rpo_pr n a c))"
      from s have "\<forall> s' \<in> set ss. size s' < size s" by (auto simp: size_simps)
      with ind have ind2: "\<And> s' t' u'. \<lbrakk>s' \<in> set ss\<rbrakk> \<Longrightarrow> ?tran s' t' u'" by blast
      with rpo_stri_imp_nstri have ind3: "\<And> us s' t' u'. \<lbrakk>s' \<in> set ss; t' \<in> set ts; u' \<in> set us\<rbrakk> \<Longrightarrow> ?tran2 s' t' u'" by blast
      {
	assume ge: "?GE s t" and gr: "?GR t u"
	have "?GR s u"
	proof (cases ?s1)
	  case True
	  from this obtain s' where s': "s' \<in> set ss" and ges: "?GE s' t" by auto
	  with s s' ind2[of s' t u, simplified] ges gr have "?GR s' u" by auto
	  hence "?GE s' u" by (rule rpo_stri_imp_nstri)
	  with s' s show ?thesis by (cases u, auto simp: Let_def)
	next
	  case False
	  show ?thesis
	  proof (cases ?t1)
	    case True
	    from this obtain t' where t': "t' \<in> set ts" and ges: "?GE t' u" by auto
	    from t' t False ge s have gr2: "?GR s t'" by (cases "\<forall> t' \<in> set ts. ?GR ?s t'", auto)
	    from t' s t gr2 ges spec[OF spec[OF spec[OF ind]], of s _ u, simplified] show "?GR s u" using size_simps supt.intros by auto
	  next
	    case False
	    from t this gr obtain h us where u: "u = Fun h us" by (cases u, auto simp: Let_def)
	    let ?u = "Fun h us"
	    from s t u ge gr have ge: "?GE ?s ?t" and gr: "?GR ?t ?u" by auto
	    from `\<not> ?s1` t ge have st': "\<forall> t' \<in> set ts. ?GR ?s t'" by (cases ?thesis, auto)
	    from `\<not> ?t1` u gr have tu': "\<forall> u' \<in> set us. ?GR ?t u'" by (cases ?thesis, auto)
	    let ?lu = "length us"
            obtain ps2 pns2 where prc2: "prc g ?lt h ?lu = (ps2,pns2)" by (cases "prc g ?lt h ?lu")
            obtain ps3 pns3 where prc3: "prc f ?ls h ?lu = (ps3,pns3)" by (cases "prc f ?ls h ?lu")
	    from `\<not> ?s1` t ge st' have fg: "pns" by (cases ?thesis, auto simp: Let_def prc) 
	    from `\<not> ?t1` u gr tu' have gh: "pns2" by (cases ?thesis, auto simp: Let_def prc2)
            note compat = prc_compat[OF prc prc2 prc3]
            from fg gh compat have fh: "pns3" by simp 
	    from tu' s t ge spec[OF spec[OF ind], of s t, simplified] u have su': "\<forall> u' \<in> set us. ?GR ?s u'" using size_simps supt.intros by auto
	    show ?thesis
	    proof (cases ps3)
	      case True
	      with su' s u fh prc3 show ?thesis by (auto)
	    next
	      case False
                from False fg gh compat have nfg: "\<not> ps" and ngh: "\<not> ps2" by blast+
                show ?thesis
                proof (cases "c f ?ls")
                case Mul note cf = this
                show ?thesis
                proof (cases "c g ?lt")
                case Mul note cg = this
                show ?thesis
                proof (cases "c h ?lu")
                case Mul note ch = this
                  from fg `\<not> ?s1` t ge st' prc nfg cf cg
                   have mul1: "snd (mul_ext (rpo_pr n) ss ts)" by auto
	          from gh `\<not> ?t1` u gr tu' prc2 ngh cg ch
                   have mul2: "fst (mul_ext (rpo_pr n) ts us)" by auto
	          from mul1 mul2 mul_ext_compat[OF ind3, of ss ts us]
                   have "fst (mul_ext (rpo_pr n) ss us)" by auto
	          with s u fh su' prc3 cf ch show ?thesis by simp
                next
                case Lex note ch = this
                  from gh `\<not> ?t1` u gr tu' prc2 ngh cg ch have us_e: "us = []" by simp
                  from gh `\<not> ?t1` u gr tu' prc2 ngh cg ch have ts_ne: "ts \<noteq> []" by simp
                  from ns_mul_ext_bottom_uniqueness[of "multiset_of ts"]
                   have "\<And>f. snd (mul_ext f [] ts) \<Longrightarrow> ts = []" unfolding mul_ext_def by (simp add: ns_mul_ext_def)
                  with ts_ne fg `\<not> ?s1` t ge st' prc nfg cf cg have ss_ne: "ss \<noteq> []"
                   by (cases ss) auto
	          from us_e ss_ne s u fh su' prc3 cf cg ch show ?thesis by simp
                qed
                next
                case Lex note cg = this
                  from fg `\<not> ?s1` t ge st' prc nfg cf cg have ts_e: "ts = []" by simp
                  with gh `\<not> ?t1` u gr tu' prc2 ngh cg show ?thesis
                   by (cases "c h ?lu") (simp_all add: lex_ext_least_2)
                qed
                next
                case Lex note cf = this 
                show ?thesis
                proof (cases "c g ?lt")
                case Mul note cg = this
                  from fg `\<not> ?s1` t ge st' prc nfg cf cg have ts_e: "ts = []" by simp
                  with gh `\<not> ?t1` u gr tu' prc2 ngh cg show ?thesis
                   by (cases "c h ?lu") (simp_all add: s_mul_ext_def s_mul_ext_bottom mul_ext_def)
                next
                case Lex note cg = this
                show ?thesis
                proof (cases "c h ?lu")
                case Mul note ch = this
                  from gh `\<not> ?t1` u gr tu' prc2 ngh cg ch have us_e: "us = []" by simp
                  from gh `\<not> ?t1` u gr tu' prc2 ngh cg ch have ts_ne: "ts \<noteq> []" by simp
                  from lex_ext_iff[of _ _ "[]" ts]
                   have "\<And>f. snd (lex_ext f n [] ts) \<Longrightarrow> ts = []" by simp
                  with ts_ne fg `\<not> ?s1` t ge st' prc nfg cf cg have ss_ne: "ss \<noteq> []"
                   by (cases ss) auto
	          from us_e ss_ne s u fh su' prc3 cf cg ch show ?thesis by simp
                next
                case Lex note ch = this
                  from fg `\<not> ?s1` t ge st' prc nfg cf cg
                   have lex1: "snd (lex_ext (rpo_pr n) n ss ts)" by auto
	         from gh `\<not> ?t1` u gr tu' prc2 ngh cg ch
                  have lex2: "fst (lex_ext (rpo_pr n) n ts us)" by auto
	         from lex1 lex2 lex_ext_compat[OF ind3, of ss ts us]
                  have "fst (lex_ext (rpo_pr n) n ss us)" by auto
	         with s u fh su' prc3 cf cg ch show ?thesis by simp
               qed
               qed
               qed
	    qed
	  qed
	qed
      }
      moreover
      {
	assume gr: "?GR s t" and ge: "?GE t u"
	have "?GR s u"
	proof (cases ?s1)
	  case True
	  from this obtain s' where s': "s' \<in> set ss" and ges: "?GE s' t" by auto
	  with s s' ind2[of s' t u, simplified] ges ge have "?GE s' u" by auto
	  with s' s show ?thesis by (cases u, auto simp: Let_def)
	next
	  case False
	  show ?thesis
	  proof (cases ?t1)
	    case True
	    from this obtain t' where t': "t' \<in> set ts" and ges: "?GE t' u" by auto
	    have tt': "?GE t t'" by (rule rpo_stri_imp_nstri[OF supt_imp_rpo_stri],
             auto simp: t t' supt.intros)
	    from t' t False gr s have gr2: "?GR s t'" by (cases "\<forall> t' \<in> set ts. ?GR ?s t'", auto)
	    from t' s t gr2 ges spec[OF spec[OF spec[OF ind]], of s _ u, simplified]
             show "?GR s u" using size_simps by auto
	  next
	    case False
	    from t this ge obtain h us where u: "u = Fun h us" by (cases u, auto simp: Let_def)
	    let ?u = "Fun h us"
	    from s t u ge gr have gr: "?GR ?s ?t" and ge: "?GE ?t ?u" by auto
	    from `\<not> ?s1` t gr have st': "\<forall> t' \<in> set ts. ?GR ?s t'" by (cases ?thesis, auto)
	    from `\<not> ?t1` u ge have tu': "\<forall> u' \<in> set us. ?GR ?t u'" by (cases ?thesis, auto)
	    let ?lu = "length us"
            obtain ps2 pns2 where prc2: "prc g ?lt h ?lu = (ps2,pns2)"
             by (cases "prc g ?lt h ?lu")
            obtain ps3 pns3 where prc3: "prc f ?ls h ?lu = (ps3,pns3)"
             by (cases "prc f ?ls h ?lu")
	    from `\<not> ?s1` t gr st' have fg: "pns" by (cases ?thesis, auto simp: Let_def prc) 
	    from `\<not> ?t1` u ge tu' have gh: "pns2" by (cases ?thesis, auto simp: Let_def prc2)
            note compat = prc_compat[OF prc prc2 prc3]
            from fg gh compat have fh: "pns3" by simp 
	    from tu' s t rpo_stri_imp_nstri[OF gr] spec[OF spec[OF ind], of s t,
              simplified] u have su': "\<forall> u' \<in> set us. ?GR ?s u'" using size_simps by auto
	    show ?thesis
	    proof (cases ps3)
	      case True
	      with su' s u fh prc3 show ?thesis by (auto)
	    next
	      case False
	      from False fg gh compat have nfg: "\<not> ps" and ngh: "\<not> ps2" by blast+
              show ?thesis
              proof (cases "c f ?ls")
              case Mul note cf = this
              show ?thesis
              proof (cases "c g ?lt")
              case Mul note cg = this
              show ?thesis
              proof (cases "c h ?lu")
              case Mul note ch = this
                from fg `\<not> ?s1` t gr st' prc nfg cf cg
                 have mul1: "fst (mul_ext (rpo_pr n) ss ts)" by auto
	        from gh `\<not> ?t1` u ge tu' prc2 ngh cg ch
                 have mul2: "snd (mul_ext (rpo_pr n) ts us)" by auto
	        from mul1 mul2 mul_ext_compat[OF ind3, of ss ts us]
                 have "fst (mul_ext (rpo_pr n) ss us)" by auto
	        with s u fh su' prc3 cf ch show ?thesis by simp
              next
              case Lex note ch = this
                from gh `\<not> ?t1` u ge tu' prc2 ngh cg ch have us_e: "us = []" by simp
                from fg `\<not> ?s1` t gr st' prc nfg cf cg s_mul_ext_bottom_strict
                 have ss_ne: "ss \<noteq> []" by (cases ss) (auto simp: s_mul_ext_def mul_ext_def)
	        from us_e ss_ne s u fh su' prc3 cf cg ch show ?thesis by simp
              qed
              next
              case Lex note cg = this
                from fg `\<not> ?s1` t gr st' prc nfg cf cg s_mul_ext_bottom_strict
                 have ss_ne: "ss \<noteq> []" by (cases ss) (auto simp: mul_ext_def)
                from fg `\<not> ?s1` t gr st' prc nfg cf cg have ts_e: "ts = []" by simp
              show ?thesis
              proof (cases "c h ?lu")
              case Mul note ch = this
                with gh `\<not> ?t1` u ge tu' prc2 ngh cg ch ns_mul_ext_bottom_uniqueness
                 have "us = []" by simp
                with ss_ne s u fh su' prc3 cf cg ch s_mul_ext_bottom
                 show ?thesis by (simp add: s_mul_ext_def mul_ext_def)
              next
              case Lex note ch = this
                from lex_ext_iff[of _ _ "[]" us]
                 have "\<And>f. snd (lex_ext f n [] us) \<Longrightarrow> us = []" by simp
                with ts_e gh `\<not> ?t1` u ge tu' prc2 ngh cg ch
                 have "us = []" by simp
                with ss_ne s u fh su' prc3 cf cg ch s_mul_ext_bottom
                 show ?thesis by (simp add: mul_ext_def)
              qed
              qed
              next
              case Lex note cf = this
              show ?thesis
              proof (cases "c g ?lt")
              case Mul note cg = this
                from fg `\<not> ?s1` t gr st' prc nfg cf cg have ss_ne: "ss \<noteq> []" by simp
                from fg `\<not> ?s1` t gr st' prc nfg cf cg have ts_e: "ts = []" by simp
              show ?thesis
              proof (cases "c h ?lu")
              case Mul note ch = this
                from ts_e gh `\<not> ?t1` u ge tu' prc2 ngh cg ch
                 ns_mul_ext_bottom_uniqueness[of "multiset_of us"]
                 have "us = []" by (simp add: s_mul_ext_def mul_ext_def)
                with ss_ne s u fh su' prc3 cf cg ch s_mul_ext_bottom
                 show ?thesis by (simp add: mul_ext_def)
              next
              case Lex note ch = this
                from gh `\<not> ?t1` u ge tu' prc2 ngh cg ch have "us = []" by simp
                with ss_ne s u fh su' prc3 cf cg ch
                 show ?thesis by (simp add: lex_ext_iff) 
              qed
              next
              case Lex note cg = this
              show ?thesis
              proof (cases "c h ?lu")
              case Mul note ch = this
                from gh `\<not> ?t1` u ge tu' prc2 ngh cg ch have us_e: "us = []" by simp
                have "\<And>f. fst (lex_ext f n ss ts) \<Longrightarrow> ss \<noteq> []"
                 by (cases ss) (simp_all add: lex_ext_iff)
                with fg `\<not> ?s1` t gr st' prc nfg cf cg have ss_ne: "ss \<noteq> []" by simp
	        with us_e s u fh su' prc3 cf cg ch show ?thesis by simp
              next
              case Lex note ch = this
	        from fg `\<not> ?s1` t gr st' prc nfg cf cg
                 have lex1: "fst (lex_ext (rpo_pr n) n ss ts)" by auto
	        from gh `\<not> ?t1` u ge tu' prc2 ngh cg ch
                 have lex2: "snd (lex_ext (rpo_pr n) n ts us)" by auto
                from lex1 lex2 lex_ext_compat[OF ind3, of ss ts us]
                 have "fst (lex_ext (rpo_pr n) n ss us)" by auto
	        with s u fh su' prc3 cf cg ch show ?thesis by auto
              qed
              qed
              qed
            qed
          qed
        qed
      }
      moreover
      {
	assume ge1: "?GE s t" and ge2: "?GE t u" and ngt1: "\<not> ?GR s t" and ngt2: "\<not> ?GR t u"
	have "\<not> ?s1"
	proof
	  assume ?s1
	  with s have "?GR s t" by (cases t, auto)
	  with ngt1 show False ..
	qed
	have "\<not> ?t1"
	proof
	  assume ?t1
	  with t have "?GR t u" by (cases u, auto)
	  with ngt2 show False ..
	qed
	from t `\<not> ?t1` ge2 obtain h us where u: "u = Fun h us" by (cases u, auto simp: Let_def)
	let ?u = "Fun h us"	    
	from s t u ge1 ge2 have ge1: "?GE ?s ?t" and ge2: "?GE ?t ?u" by auto
	from `\<not> ?s1` t ge1 have st': "\<forall> t' \<in> set ts. ?GR ?s t'" by (cases ?thesis, auto)
	from `\<not> ?t1` u ge2 have tu': "\<forall> u' \<in> set us. ?GR ?t u'" by (cases ?thesis, auto)
	let ?lu = "length us"        
        obtain ps2 pns2 where prc2: "prc g ?lt h ?lu = (ps2,pns2)" by (cases "prc g ?lt h ?lu")
        obtain ps3 pns3 where prc3: "prc f ?ls h ?lu = (ps3,pns3)" by (cases "prc f ?ls h ?lu")
	from `\<not> ?s1` t ge1 st' have fg: "pns" by (cases ?thesis, auto simp: Let_def prc) 
	from `\<not> ?t1` u ge2 tu' have gh: "pns2" by (cases ?thesis, auto simp: Let_def prc2)
        note compat = prc_compat[OF prc prc2 prc3]
        from compat fg gh have fh: pns3 by blast
	from tu' s t ge1 spec[OF spec[OF ind], of s t, simplified] u
         have su': "\<forall> u' \<in> set us. ?GR ?s u'" using size_simps by auto
	from `\<not> ?s1` st' ge1 ngt1 s t have nfg: "\<not> ps"
         by (simp add: Let_def, cases ?thesis, simp, cases ps, auto simp: prc fg) 
	from `\<not> ?t1` tu' ge2 ngt2 t u have ngh: "\<not> ps2"
         by (simp add: Let_def, cases ?thesis, simp, cases ps2, auto simp: prc2 gh)
        have "?GE s u"
        proof (cases "c f ?ls")
        case Mul note cf = this
        show ?thesis
        proof (cases "c g ?lt")
        case Mul note cg = this
        show ?thesis
        proof (cases "c h ?lu")
        case Mul note ch = this
          from fg `\<not> ?s1` t ge1 st' prc nfg cf cg
           have mul1: "snd (mul_ext (rpo_pr n) ss ts)" by auto
	  from gh `\<not> ?t1` u ge2 tu' prc2 ngh cg ch
           have mul2: "snd (mul_ext (rpo_pr n) ts us)" by auto
          from mul1 mul2 mul_ext_compat[OF ind3, of ss ts us]
           have "snd (mul_ext (rpo_pr n) ss us)" by auto
          with s u fh su' prc3 cf ch show ?thesis by simp
        next
        case Lex note ch = this
          from gh `\<not> ?t1` u ge2 tu' prc2 ngh cg ch have us_e: "us = []" by simp
          with s u fh su' prc3 cf cg ch show ?thesis by simp
        qed
        next
        case Lex note cg = this
          from fg `\<not> ?s1` t ge1 st' prc nfg cf cg have ts_e: "ts = []" by simp
        show ?thesis
        proof (cases "c h ?lu")
        case Mul note ch = this
          with gh `\<not> ?t1` u ge2 tu' prc2 ngh cg ch have "us = []" by simp
          with s u fh su' prc3 cf cg ch ns_mul_ext_bottom
           show ?thesis by (simp add: ns_mul_ext_def mul_ext_def)
        next
        case Lex note ch = this
          have "\<And>f. snd (lex_ext f n [] us) \<Longrightarrow> us = []" by (simp_all add: lex_ext_iff)
          with ts_e gh `\<not> ?t1` u ge2 tu' prc2 ngh cg ch have "us = []" by simp
          with s u fh su' prc3 cf cg ch show ?thesis by simp
        qed
        qed
        next
        case Lex note cf = this
        show ?thesis
        proof (cases "c g ?lt")
        case Mul note cg = this
          from fg `\<not> ?s1` t ge1 st' prc nfg cf cg have ts_e: "ts = []" by simp
        show ?thesis
        proof (cases "c h ?lu")
        case Mul note ch = this
          with ts_e gh `\<not> ?t1` u ge2 tu' prc2 ngh cg ch
           ns_mul_ext_bottom_uniqueness[of "multiset_of us"]
           have "us = []" by (simp add: s_mul_ext_def mul_ext_def)
          with s u fh su' prc3 cf cg ch show ?thesis by simp
        next
        case Lex note ch = this
          with gh `\<not> ?t1` u ge2 tu' prc2 ngh cg ch have "us = []" by simp
          with s u fh su' prc3 cf cg ch show ?thesis by (simp add: lex_ext_least_1)
        qed
        next
        case Lex note cg = this
        show ?thesis
        proof (cases "c h ?lu")
        case Mul note ch = this
          with gh `\<not> ?t1` u ge2 tu' prc2 ngh cg ch have "us = []" by simp
          with s u fh su' prc3 cf cg ch show ?thesis by (simp add: lex_ext_least_1)
        next
        case Lex note ch = this
          from `\<not> ?s1` st' ge1 s t fg nfg cf cg
           have lex1: "snd (lex_ext (rpo_pr n) n ss ts)" by (auto simp: prc)
          from `\<not> ?t1` tu' ge2 t u gh ngh cg ch
           have lex2: "snd (lex_ext (rpo_pr n) n ts us)" by (auto simp: prc2)
          from lex1 lex2 lex_ext_compat[OF ind3, of ss ts us]
          have "snd (lex_ext (rpo_pr n) n ss us)" by auto
          with fg gh su' s u fh cf cg ch show ?thesis by (auto simp: prc3)
        qed
        qed
        qed
      }
      ultimately
      show ?thesis using rpo_stri_imp_nstri[of "pr" c n s u] by auto
    qed
  qed
qed

(* Transitivity of the strict order *)

lemma rpo_trans:
  assumes one: "fst (rpo_pr n s t)"
  and two: "fst (rpo_pr n t u)"
  shows "fst (rpo_pr n s u)"
using assms rpo_compat rpo_stri_imp_nstri[OF one] by blast

(* large_symbol implies orientation *)
lemma rpo_large_sym:
  assumes prec: "\<And> g m. (g,m) \<in> funas_term t \<Longrightarrow> prc f (length ss) g m = (True,True)"
  and vars: "vars_term t \<subseteq> vars_term (Fun f ss)"
  shows "rpo_pr n (Fun f ss) t = (True,True)"
  using assms
proof (induct t)
  case (Var x)
  hence "x \<in> vars_term (Fun f ss)" by simp
  from supteq_Var[OF this] 
  have"Fun f ss \<rhd> Var x" unfolding subterm.le_less by simp
  from supt_imp_rpo_stri[OF this] have "fst (rpo_pr n (Fun f ss) (Var x))" .
  with rpo_stri_imp_nstri[OF this] show ?case by (cases "rpo_pr n (Fun f ss) (Var x)", auto)
next
  case (Fun g ts)
  from Fun(2) have prec: "prc f (length ss) g (length ts) = (True,True)" by simp
  { 
    fix t
    assume t: "t \<in> set ts"
    from t Fun(3) have v: "vars_term t \<subseteq> vars_term (Fun f ss)" by auto
    from t Fun(2) have p: "\<And> gg m. (gg, m) \<in> funas_term t \<Longrightarrow> prc f (length ss) gg m = (True,True)"
     by auto
    from Fun(1)[OF t p v] have "rpo_pr n (Fun f ss) t = (True,True)" .
  }
  thus ?case
    by (simp add: prec)
qed

(* Definition: corresponding relations *)
definition RPO_S_pr where "RPO_S_pr n = {(s,t) . fst (rpo_pr n s t)}"
definition RPO_NS_pr where "RPO_NS_pr n = {(s,t). snd (rpo_pr n s t)}"

(* introduction rules w.r.t. inference rules *)
lemma RPO_subt: "s \<in> set ss \<Longrightarrow> (s,t) \<in> RPO_NS_pr n \<Longrightarrow> (Fun f ss,t) \<in> RPO_S_pr n"
  unfolding RPO_S_pr_def RPO_NS_pr_def by (cases t, auto simp: Let_def)

lemma RPO_prec: "\<lbrakk>\<And> t. t \<in> set ts \<Longrightarrow> (Fun f ss, t) \<in> RPO_S_pr n\<rbrakk> \<Longrightarrow> 
  fst (prc f (length ss) g (length ts)) \<Longrightarrow> (Fun f ss, Fun g ts) \<in> RPO_S_pr n"
  unfolding RPO_S_pr_def 
  using prc_stri_imp_nstri[of f "length ss" g "length ts"]
  by (cases "prc f (length ss) g (length ts)", auto)

lemma RPO_refl: "(s,s) \<in> RPO_NS_pr n"
  unfolding RPO_NS_pr_def 
  by (simp add: rpo_nstri_refl)

lemma RPO_stri: "(s,t) \<in> RPO_S_pr n \<Longrightarrow> (s,t) \<in> RPO_NS_pr n" 
  unfolding RPO_S_pr_def RPO_NS_pr_def 
  using rpo_stri_imp_nstri by auto

lemma RPO_least_Var: "prl g \<Longrightarrow> (Var x,Fun g []) \<in> RPO_NS_pr n"
  unfolding RPO_NS_pr_def   
  using rpo_least_1 by blast

lemma RPO_least_Fun: "prl g \<Longrightarrow> (Fun f ss,Fun g []) \<in> RPO_NS_pr n"
  unfolding RPO_NS_pr_def   
  using rpo_least_1 by blast

lemma RPO_S_least_Fun: "prl g \<Longrightarrow> ss \<noteq> [] \<Longrightarrow> (Fun f ss, Fun g []) \<in> RPO_S_pr n"
  by (rule RPO_subt[where s = "hd ss"], simp, unfold RPO_NS_pr_def, 
    insert rpo_least_1, blast)

lemma RPO_S_mul: assumes ms: "(multiset_of ss, multiset_of ts) \<in> s_mul_ext (RPO_NS_pr n) (RPO_S_pr n)"
  and prc: "snd (prc f (length ss) g (length ts))"
  and fmul: "c f (length ss) = Mul"
  and gmul: "c g (length ts) = Mul"
  shows "(Fun f ss, Fun g ts) \<in> RPO_S_pr n"
  unfolding RPO_S_pr_def
proof (rule, unfold split)
  let ?n = "length ss"
  let ?m = "length ts"
  from prc obtain b where prc: "prc f ?n g ?m = (b,True)" by (cases "prc f ?n g ?m", auto)
  from ms have ms': "fst (mul_ext (rpo_pr n) ss ts)"
    unfolding mul_ext_def Let_def RPO_NS_pr_def RPO_S_pr_def by simp
  from ms[unfolded s_mul_ext_def]
  obtain as bs A' B' where 
    id: "multiset_of ss = multiset_of as + A'"
    "multiset_of ts = multiset_of bs + B'"
    and len: "length as = length bs"
    and as: "\<And> i. i < length bs \<Longrightarrow> (as ! i, bs ! i) \<in> RPO_NS_pr n"
    and A': "\<And> b. b \<in># B' \<Longrightarrow> \<exists> a. a \<in># A' \<and> (a,b) \<in> RPO_S_pr n" by blast
  {
    fix t
    assume "t \<in> set ts"
    hence "t \<in># multiset_of ts" unfolding in_multiset_in_set .
    hence "t \<in># multiset_of bs \<or> t \<in># B'" unfolding id by simp
    hence "\<exists> s. s \<in># multiset_of as + A' \<and> (s,t) \<in> RPO_NS_pr n"
    proof
      assume "t \<in># B'"
      from A'[OF this] obtain s where A': "s \<in># A'" and S: "(s,t) \<in> RPO_S_pr n"
        by blast
      from RPO_stri[OF S] A' show ?thesis by auto
    next
      assume "t \<in># multiset_of bs"
      hence "t \<in> set bs" unfolding in_multiset_in_set .
      from this[unfolded set_conv_nth] obtain i where t: "t = bs ! i"
        and i: "i < length bs" by auto
      from as[OF i] have NS: "(as ! i, t) \<in> RPO_NS_pr n" unfolding t .
      from i len have "as ! i \<in># multiset_of as" 
        unfolding in_multiset_in_set set_conv_nth by auto
      with NS show ?thesis by auto
    qed
    then obtain s where s: "s \<in> set ss" and NS: "(s,t) \<in> RPO_NS_pr n"
      unfolding in_multiset_in_set id[symmetric] by blast
    from RPO_subt[OF this] have "(Fun f ss, t) \<in> RPO_S_pr n" .
  }
  thus "fst (rpo_pr n (Fun f ss) (Fun g ts))"
    unfolding RPO_S_pr_def
    by (auto simp: fmul gmul prc ms')
qed

lemma RPO_NS_mul: assumes ms: "(multiset_of ss, multiset_of ts) \<in> ns_mul_ext (RPO_NS_pr n) (RPO_S_pr n)"
  and prc: "snd (prc f (length ss) g (length ts))"
  and fmul: "c f (length ss) = Mul"
  and gmul: "c g (length ts) = Mul"
  shows "(Fun f ss, Fun g ts) \<in> RPO_NS_pr n"
  unfolding RPO_NS_pr_def
proof (rule, unfold split)
  let ?n = "length ss"
  let ?m = "length ts"
  from prc obtain b where prc: "prc f ?n g ?m = (b,True)" by (cases "prc f ?n g ?m", auto)
  from ms have ms': "snd (mul_ext (rpo_pr n) ss ts)"
    unfolding mul_ext_def Let_def RPO_NS_pr_def RPO_S_pr_def by simp
  from ms[unfolded ns_mul_ext_def]
  obtain as bs A' B' where 
    id: "multiset_of ss = multiset_of as + A'"
    "multiset_of ts = multiset_of bs + B'"
    and len: "length as = length bs"
    and as: "\<And> i. i < length bs \<Longrightarrow> (as ! i, bs ! i) \<in> RPO_NS_pr n"
    and A': "\<And> b. b \<in># B' \<Longrightarrow> \<exists> a. a \<in># A' \<and> (a,b) \<in> RPO_S_pr n" by blast
  {
    fix t
    assume "t \<in> set ts"
    hence "t \<in># multiset_of ts" unfolding in_multiset_in_set .
    hence "t \<in># multiset_of bs \<or> t \<in># B'" unfolding id by simp
    hence "\<exists> s. s \<in># multiset_of as + A' \<and> (s,t) \<in> RPO_NS_pr n"
    proof
      assume "t \<in># B'"
      from A'[OF this] obtain s where A': "s \<in># A'" and S: "(s,t) \<in> RPO_S_pr n"
        by blast
      from RPO_stri[OF S] A' show ?thesis by auto
    next
      assume "t \<in># multiset_of bs"
      hence "t \<in> set bs" unfolding in_multiset_in_set .
      from this[unfolded set_conv_nth] obtain i where t: "t = bs ! i"
        and i: "i < length bs" by auto
      from as[OF i] have NS: "(as ! i, t) \<in> RPO_NS_pr n" unfolding t .
      from i len have "as ! i \<in># multiset_of as" 
        unfolding in_multiset_in_set set_conv_nth by auto
      with NS show ?thesis by auto
    qed
    then obtain s where s: "s \<in> set ss" and NS: "(s,t) \<in> RPO_NS_pr n"
      unfolding in_multiset_in_set id[symmetric] by blast
    from RPO_subt[OF this] have "(Fun f ss, t) \<in> RPO_S_pr n" .
  }
  thus "snd (rpo_pr n (Fun f ss) (Fun g ts))"
    unfolding RPO_S_pr_def
    by (auto simp: fmul gmul prc ms')
qed

lemma RPO_lex: 
  defines lexext: "lexext n \<equiv> lex_ext (\<lambda> a b. ((a,b) \<in> RPO_S_pr n,(a,b) \<in> RPO_NS_pr n)) n"
  assumes lex: "fst (lexext n ss ts)"
  and prc: "snd (prc f (length ss) g (length ts))"
  and ts: "\<And> t. t \<in> set ts \<Longrightarrow> (Fun f ss,t) \<in> RPO_S_pr n"
  and fmul: "c f (length ss) = Lex"
  and gmul: "c g (length ts) = Lex"
  shows "(Fun f ss, Fun g ts) \<in> RPO_S_pr n"
  unfolding RPO_S_pr_def
proof (rule, unfold split)
  let ?n = "length ss"
  let ?m = "length ts"
  from prc obtain b where prc: "prc f ?n g ?m = (b,True)" by (cases "prc f ?n g ?m", auto)
  note lex = lex[unfolded lexext RPO_NS_pr_def RPO_S_pr_def] 
  with ts[unfolded RPO_S_pr_def]
  show "fst (rpo_pr n (Fun f ss) (Fun g ts))"
    unfolding RPO_S_pr_def
    by (auto simp: fmul gmul prc)
qed

lemma RPO_NS_lex: 
  defines lexext: "lexext n \<equiv> lex_ext (\<lambda> a b. ((a,b) \<in> RPO_S_pr n,(a,b) \<in> RPO_NS_pr n)) n"
  assumes lex: "snd (lexext n ss ts)"
  and prc: "snd (prc f (length ss) g (length ts))"
  and ts: "\<And> t. t \<in> set ts \<Longrightarrow> (Fun f ss,t) \<in> RPO_S_pr n"
  and fmul: "c f (length ss) = Lex"
  and gmul: "c g (length ts) = Lex"
  shows "(Fun f ss, Fun g ts) \<in> RPO_NS_pr n"
  unfolding RPO_NS_pr_def
proof (rule, unfold split)
  let ?n = "length ss"
  let ?m = "length ts"
  from prc obtain b where prc: "prc f ?n g ?m = (b,True)" by (cases "prc f ?n g ?m", auto)
  note lex = lex[unfolded lexext RPO_NS_pr_def RPO_S_pr_def] 
  with ts[unfolded RPO_S_pr_def]
  show "snd (rpo_pr n (Fun f ss) (Fun g ts))"
    unfolding RPO_S_pr_def
    by (auto simp: fmul gmul prc)
qed


(* desired properties of RPO *)
lemma RPO_NS_refl: "refl (RPO_NS_pr n)"
unfolding refl_on_def RPO_NS_pr_def
 by (simp add: rpo_nstri_refl)

lemma RPO_NS_trans: "trans (RPO_NS_pr n)"
unfolding trans_def RPO_NS_pr_def 
using rpo_compat by blast

lemma RPO_S_trans: "trans (RPO_S_pr n)"
unfolding trans_def RPO_S_pr_def 
using rpo_trans by blast

lemma RPO_compat: "RPO_NS_pr n O RPO_S_pr n \<subseteq> RPO_S_pr n"
  using rpo_compat[of n]
  by (unfold RPO_S_pr_def RPO_NS_pr_def, auto)


(* Strong Normalisation *)

lemma RPO_S_SN: "SN (RPO_S_pr m :: ('f,'v)term rel)"
unfolding RPO_S_pr_def
proof - {
  fix t :: "('f,'v)term"
  let ?Rp = "\<lambda> s t. fst (rpo_pr m s t)"
  let ?R = "{(s,t). ?Rp s t}"
  let ?S = "\<lambda>x. SN_on ?R {x}"
  note iff = SN_on_all_reducts_SN_on_conv[of ?R]
  {
    fix x
    have "?S (Var x)" unfolding iff[of "Var x"]
    proof (intro allI impI)
      fix s
      assume "(Var x, s) \<in> ?R"
      hence False by (cases s, auto)
      thus "?S s" ..
    qed
  } note var_SN = this
  have "?S t"
  proof (induct t)
  case (Var x) show ?case by (rule var_SN)
  next
  case (Fun f ts)
    let ?Slist = "\<lambda> ys. \<forall> y \<in> set ys. ?S y"
    let ?cond = "\<lambda> ((f,k),ab) ((g,l),ab'). ((c f k = c g l) \<longrightarrow> (?Slist ab \<and>
      (c f k = Mul \<longrightarrow> fst (mul_ext (rpo_pr m) ab ab')) \<and>
      (c f k = Lex \<longrightarrow> fst (lex_ext (rpo_pr m) m ab ab'))))
     \<and> ((c f k \<noteq> c g l) \<longrightarrow> (ab \<noteq> [] \<and> ab' = []))"
    let ?r3 = "{(ab, ab'). ?cond ab ab'}"
    let ?r1 = "lex_two (\<lambda> (f,n) (g,m). fst (prc f n g m)) (\<lambda> (f,n) (g,m).
     snd (prc f n g m)) ?cond"
    let ?ge = "\<lambda> t s. snd (rpo_pr m t s)"
    let ?gr = ?Rp
    {
      fix ab
      {
        assume "\<exists>S. S 0 = ab \<and> (\<forall>i. (S i, S (Suc i)) \<in> ?r3)"
        then obtain S where
          S0: "S 0 = ab" and
          SS: "\<forall>i. (S i, S (Suc i)) \<in> ?r3"
        by auto
        let ?Sf = "\<lambda>i. fst (fst (S i))"
        let ?Sn = "\<lambda>i. snd (fst (S i))"
        let ?Sts = "\<lambda>i. snd (S i)"
        have False
        proof (cases "\<forall>i. c (?Sf i) (?Sn i) = Mul")
        case True
          let ?r' = "{(ys, xs).
           (\<forall>y\<in>set ys. SN_on {(s, t). fst (local.rpo_pr m s t)} {y})
           \<and> fst (mul_ext (rpo_pr m) ys xs)}"
          {
            fix i
            from True[rule_format, of i] and True[rule_format, of "Suc i"]
             and SS[rule_format, of i]
             have "(snd (S i), (snd (S (Suc i)))) \<in> ?r'" by auto
          }
          hence Hf: "\<not> SN_on ?r' {snd (S 0)}"
           unfolding SN_on_def by auto
          from mul_ext_SN[of "rpo_pr m", rule_format, OF rpo_nstri_refl]
            and rpo_compat[of m] and rpo_stri_imp_nstri[of "pr" c m]
            have "SN ?r'" by blast
          from SN_on_subset2[OF subset_UNIV[of "{snd (S 0)}"], OF this]
            have "SN_on ?r' {snd (S 0)}" .
          with Hf show ?thesis ..
        next
        case False note HMul = this
        show ?thesis
        proof (cases "\<forall>i. c (fst (fst (S i))) (snd (fst (S i))) = Lex")
        case True
          let ?r' = "{(ys, xs).
           (\<forall>y\<in>set ys. SN_on {(s, t). fst (local.rpo_pr m s t)} {y})
           \<and> fst (lex_ext (rpo_pr m) m ys xs)}"
          {
            fix i
            from True[rule_format, of i] and True[rule_format, of "Suc i"]
             and SS[rule_format, of i]
             have "(snd (S i), (snd (S (Suc i)))) \<in> ?r'" by auto
          }
          hence Hf: "\<not> SN_on ?r' {snd (S 0)}"
           unfolding SN_on_def by auto
          from lex_ext_SN[of "rpo_pr m" m] and rpo_stri_imp_nstri and rpo_compat[of m]
            have "SN ?r'" by blast
          from SN_on_subset2[OF subset_UNIV[of "{snd (S 0)}"] this]
            have "SN_on ?r' {snd (S 0)}" .
          with Hf show ?thesis ..
        next
        case False note HLex = this
          from HMul and HLex
           have "\<exists>i. c (fst (fst (S i))) (snd (fst (S i)))
                 \<noteq> c (fst (fst (S (Suc i)))) (snd (fst (S (Suc i))))"
          proof (cases ?thesis, simp)
          case False
            hence T: "\<forall>i. c (fst (fst (S i))) (snd (fst (S i)))
                 = c (fst (fst (S (Suc i)))) (snd (fst (S (Suc i)))) " by simp
            {
              fix i
              have "c (fst (fst (S i))) (snd (fst (S i)))
                  = c (fst (fst (S 0))) (snd (fst (S 0)))"
              proof (induct i)
              case 0 thus ?case by simp
              next
              case (Suc j) thus ?case by (simp add: T[rule_format, of j])
              qed
            }
            thus ?thesis using HMul HLex
             by (cases "c (fst (fst (S 0))) (snd (fst (S 0)))") auto
          qed
          then obtain i where
            Hdiff: "c (fst (fst (S i))) (snd (fst (S i)))
                  \<noteq> c (fst (fst (S (Suc i)))) (snd (fst (S (Suc i))))"
           by auto
           from Hdiff have Hf: "snd (S (Suc i)) = []"
            using SS[rule_format, of i] by auto
          show ?thesis
          proof (cases "c (fst (fst (S (Suc i)))) (snd (fst (S (Suc i))))
          = c (fst (fst (S (Suc (Suc i))))) (snd (fst (S (Suc (Suc i)))))")
          case False thus ?thesis using Hf and SS[rule_format, of "Suc i"] by auto
          next
          case True
            show ?thesis
            proof (cases "c (fst (fst (S (Suc i)))) (snd (fst (S (Suc i))))")
            case Mul
              with True and SS[rule_format, of "Suc i"]
               have "fst (mul_ext (rpo_pr m) (snd (S (Suc i)))
                 (snd (S (Suc (Suc i)))))"
               by auto
              with Hf and s_mul_ext_bottom_strict show ?thesis
               by (simp add: s_mul_ext_def mul_ext_def)
            next
            case Lex
              with True and SS[rule_format, of "Suc i"]
               have "fst (lex_ext (rpo_pr m) m (snd (S (Suc i)))
                 (snd (S (Suc (Suc i)))))"
               by auto
              with Hf show ?thesis by (simp add: lex_ext_iff)
            qed
          qed
        qed
        qed
      }
      hence "SN_on ?r3 {ab}" unfolding SN_on_def unfolding singleton_iff ..
    }
    hence "SN ?r3" unfolding SN_on_def by blast
    have SN:"SN {(ab,ab'). ?r1 ab ab'}" (is "SN ?r0")
    proof (rule lex_two[OF _ _ `SN ?r3`], clarify)
      fix a1 b1 a2 b2 a3 b3
      assume ns: "snd (prc a1 b1 a2 b2)" and s: "fst (prc a2 b2 a3 b3)"
      show "fst (prc a1 b1 a3 b3)" 
       by (cases "prc a1 b1 a2 b2", cases "prc a2 b2 a3 b3", cases "prc a1 b1 a3 b3", 
        insert ns s prc_compat[of a1 b1 a2 b2 _ _ a3 b3, THEN conjunct2, THEN conjunct1],
        force)
    next
      let ?r'' = "{(a,a'). (\<lambda>(f,n) (g,m). fst (prc f n g m)) a a'}"
      let ?r' = "{((f, n), g, m) |f n g m. fst (prc f n g m)}"
      have "?r'' = ?r'" by auto
      thus "SN ?r''" using prc_SN by auto
    qed
    let ?m = "\<lambda> (f,ts). ((f,length ts), ((f, length ts), ts))"
    let ?r = "{(a,b). (?m a, ?m b) \<in> ?r0}"
    have SN_r: "SN ?r" using SN_inv_image[OF SN, of ?m] unfolding inv_image_def by fast
    let ?ind = "\<lambda> (f,ts). ?Slist ts \<longrightarrow> ?S (Fun f ts)"
    have "?ind (f,ts)"
    proof (rule SN_induct[OF SN_r, of ?ind])
      fix fts
      assume ind: "\<And> gss. (fts,gss) \<in> ?r \<Longrightarrow> ?ind gss"
      show "?ind fts"
      proof (cases fts)
      case (Pair f ts)
        thus ?thesis
        proof (simp, intro impI)
	  assume ts: "?Slist ts"
          let ?t = "Fun f ts"	
          show "?S ?t"
          proof (simp only: iff[of ?t], intro allI impI)
            fix s
	    assume "(?t,s) \<in> ?R"
            hence "?gr ?t s" by simp
	    thus "?S s"
	    proof (induct s, simp add: var_SN)
	    case (Fun g ss) note IH = this
              let ?s = "Fun g ss"
              from Fun have t_gr_s: "?gr ?t ?s" by auto
 	      show "?S ?s"
	      proof (cases "\<exists> t \<in> set ts. ?ge t ?s")
	      case True
	        then obtain t where "t \<in> set ts" and ge: "?ge t ?s" by auto	      
	        with ts have "?S t" by auto
                show "?S ?s"
                proof (unfold iff[of ?s], intro allI impI)
                  fix u
                  assume "(?s,u) \<in> ?R"
                  with rpo_compat ge have u: "?gr t u" by blast
                  with `?S t`[unfolded iff[of t]]
                  show "?S u" by simp
                qed
	      next
	      case False note oFalse = this
	        let ?ls = "length ss"
	        let ?lt = "length ts"
                obtain s1 ns1 where prc1: "prc f ?lt g ?ls = (s1,ns1)"
                  by (cases "prc f ?lt g ?ls")
	        from False t_gr_s have f_ge_g: "ns1"
                  by (cases ?thesis, auto simp: Let_def prc1)
	        from False t_gr_s f_ge_g have small_ss: "\<forall> s \<in> set ss. ?gr ?t s"
                  by (cases ?thesis, auto simp: Let_def)
	        with Fun have ss_S: "?Slist ss" by auto
	        show ?thesis
	        proof (cases s1)
		case True
		  hence "((f,ts),(g,ss)) \<in> ?r"  by (simp add: prc1)
		  with ind have "?ind (g,ss)" using Pair by auto
		  with ss_S show ?thesis by auto
	        next
		case False
                  show ?thesis
                  proof (cases "c f ?lt")
                  case Mul note cf = this
                  show ?thesis
                  proof (cases "c g ?ls")
                  case Mul note cg = this
		    from False oFalse t_gr_s small_ss f_ge_g cf cg
                    have mul: "fst (mul_ext (rpo_pr m) ts ss)"
                     by (cases ?thesis, auto simp: Let_def prc1)
		    from False mul ts f_ge_g cf cg have "((f,ts),(g,ss)) \<in> ?r"
                     by (simp add: prc1)
                    with ind have "?ind (g,ss)" using Pair by auto
		    with ss_S show ?thesis by auto
                  next
                  case Lex note cg = this
                    from cf cg False oFalse f_ge_g t_gr_s small_ss prc1
                    have "((f,ts),(g,ss)) \<in> ?r" by auto
  	            with ind have "?ind (g,ss)" using Pair by auto
		    with ss_S show ?thesis by simp
                  qed
                  next
                  case Lex note cf = this
                  show ?thesis
                  proof (cases "c g ?ls")
                  case Mul note cg = this
                    from cf cg False oFalse f_ge_g t_gr_s small_ss prc1
                    have "((f,ts),(g,ss)) \<in> ?r" by auto
  	            with ind have "?ind (g,ss)" using Pair by auto
		    with ss_S show ?thesis by simp
                  next
                  case Lex note cg = this
		    from False oFalse t_gr_s small_ss f_ge_g cf cg
                    have mul: "fst (lex_ext (rpo_pr m) m ts ss)"
                     by (cases ?thesis, auto simp: Let_def prc1)
		    from False mul ts f_ge_g cf cg have "((f,ts),(g,ss)) \<in> ?r"
                     by (simp add: prc1)
                    with ind have "?ind (g,ss)" using Pair by auto
		    with ss_S show ?thesis by auto
                  qed
                qed
              qed
            qed
          qed
        qed
      qed
    qed
  qed
  with Fun show ?case by simp
qed
}
  from SN_I[OF this]
    show "SN {(s::('f, 'v)term, t). fst (rpo_pr m s t)}" .
qed

end

(* rpo as reduction pairs *)

sublocale rpo_pr \<subseteq> rpo_redpair: mono_ce_af_redtriple "RPO_S_pr n" "RPO_NS_pr n" "RPO_NS_pr n" full_af
proof (unfold_locales, rule RPO_S_SN)
  show "subst_closed (RPO_NS_pr n)"
    unfolding subst_closed_def 
  proof (rule subrelI)
    fix s t :: "('f, 'v) term"
    assume "(s,t) \<in> CS (RPO_NS_pr n)"
    then obtain u v and \<sigma>::"('f, 'v) subst"
      where "(u,v) \<in> RPO_NS_pr n" and "u\<cdot>\<sigma> = s" and "v\<cdot>\<sigma> = t"
      by (cases rule: subst_closure.cases) auto
    thus "(s,t) \<in> RPO_NS_pr n" unfolding RPO_NS_pr_def using rpo_stable[of n] by auto
  qed
next
  show "subst_closed (RPO_S_pr n)"
    unfolding subst_closed_def 
  proof (rule subrelI)
    fix s t :: "('f, 'v) term"
    assume "(s,t) \<in> CS (RPO_S_pr n)"
    then obtain u v and \<sigma> :: "('f, 'v) subst"
      where "(u,v) \<in> RPO_S_pr n" and "u\<cdot>\<sigma> = s" and "v\<cdot>\<sigma> = t"
      by (cases rule: subst_closure.cases) auto
    thus "(s,t) \<in> RPO_S_pr n" unfolding RPO_S_pr_def using rpo_stable[of n] by auto
  qed
next
  show "ctxt_closed (RPO_NS_pr n)"
    by (rule one_imp_ctxt_closed, unfold RPO_NS_pr_def, insert rpo_nstri_mono, blast)
next
  show "\<exists> k. \<forall> m \<ge> k. \<forall> c d. ce_trs (c,d,m) \<subseteq> RPO_NS_pr n"
  proof (intro exI allI impI)
    fix m c d
    show "ce_trs (c,d,m) \<subseteq> RPO_NS_pr n"
      unfolding RPO_NS_pr_def 
      by (simp add: ce_trs.simps, auto simp: size_simps supt.intros rpo_stri_imp_nstri[OF supt_imp_rpo_stri])
  qed
next
  show "\<exists> k. \<forall> m \<ge> k. \<forall> c d. ce_trs (c,d,m) \<subseteq> RPO_S_pr n"
  proof (intro exI allI impI)
    fix m c d
    show "ce_trs (c,d,m) \<subseteq> RPO_S_pr n"
      unfolding RPO_S_pr_def 
      by (simp add: ce_trs.simps, auto simp: size_simps supt.intros supt_imp_rpo_stri)
  qed
next
  show "RPO_S_pr n O RPO_NS_pr n \<subseteq> RPO_S_pr n"
    unfolding RPO_S_pr_def RPO_NS_pr_def using rpo_compat[of n] by auto
next
qed (rule RPO_compat, rule full_af)

sublocale rpo_pr \<subseteq> rpo_redpair: redtriple_order "RPO_S_pr n" "RPO_NS_pr n" "RPO_NS_pr n" 
  by (unfold_locales, insert RPO_NS_refl RPO_NS_trans RPO_S_trans, auto)
  

context rpo_pr
begin

lemma RPO_S_pr_ctxt_closed: "ctxt_closed (RPO_S_pr n)"
  by (rule one_imp_ctxt_closed, unfold RPO_S_pr_def, insert rpo_stri_mono, blast)

lemma rpo_manna_ness: assumes "\<And> l r. (l,r) \<in> R \<Longrightarrow> fst (rpo_pr n l r)"
  shows "SN (rstep R)"
  by (rule manna_ness, rule RPO_S_pr_ctxt_closed[of n], insert assms, unfold RPO_S_pr_def, auto)
end

(* Definitions RPO with precedence via natural numbers *)

definition prc_nat :: "('f \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> ('f \<Rightarrow> nat \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> bool \<times> bool)"
  where "prc_nat pr \<equiv> \<lambda> f n g m. let pf = pr f n; pg = pr g m in (pg < pf, pg \<le> pf)"

definition prl_nat :: "('f \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> 'f \<Rightarrow> bool"
  where "prl_nat pr \<equiv> \<lambda> f. pr f 0 = 0"

abbreviation pr_nat where "pr_nat pr \<equiv> (prc_nat pr, prl_nat pr)"


definition
  rpo_strict 
where
  "rpo_strict pr \<tau> n \<equiv> \<lambda>(s, t). check (fst (rpo (pr_nat pr) \<tau> n s t)) (''could not orient '' +#+ shows s +@+ '' >RPO '' +#+ shows t +@+ shows_nl)"

definition
  rpo_nstrict
where
  "rpo_nstrict pr \<tau> n \<equiv> \<lambda> (s,t). check (snd (rpo (pr_nat pr) \<tau> n s t)) (''could not orient '' +#+ shows s +@+ '' >=RPO '' +#+ shows t +@+ shows_nl)"

lemma SN_pr_nat: "SN {((f,n), g, m) | f n g m. (pr :: 'f \<Rightarrow> nat \<Rightarrow> nat) g m < pr f n}"
  by (rule HOL.subst[of _ _ SN, OF _ SN_inv_image[OF SN_nat_gt, of "\<lambda> (f,n). pr f n"]],
  auto simp: inv_image_def)

interpretation rpo_pr_prc: rpo_pr "prc_nat pr" "prl_nat pr" c
  for "pr" :: "'f \<Rightarrow> nat \<Rightarrow> nat" and c :: "'f \<Rightarrow> nat \<Rightarrow> order_tag"
  by (unfold_locales, unfold prc_nat_def prl_nat_def Let_def, auto simp: SN_pr_nat)

abbreviation RPO_S where "RPO_S pr \<tau> n \<equiv> rpo_pr.RPO_S_pr (prc_nat pr) (prl_nat pr) \<tau> n"
abbreviation RPO_NS where "RPO_NS pr \<tau> n \<equiv> rpo_pr.RPO_NS_pr (prc_nat pr) (prl_nat pr) \<tau> n"

interpretation RPO_redtriple_impl: redtriple_impl "RPO_S pr \<tau> n" "RPO_NS pr \<tau> n" "RPO_NS pr \<tau> n" "rpo_strict pr \<tau> n" "rpo_nstrict pr \<tau> n" "rpo_nstrict pr \<tau> n"
  by (unfold_locales, unfold rpo_strict_def rpo_nstrict_def rpo_pr_prc.RPO_S_pr_def rpo_pr_prc.RPO_NS_pr_def, auto)

interpretation RPO_ce_mono_redpair: mono_ce_af_redtriple_order_impl "RPO_S pr \<tau> n" "RPO_NS pr \<tau> n" "RPO_NS pr \<tau> n" "rpo_strict pr \<tau> n" "rpo_nstrict pr \<tau> n" "rpo_nstrict pr \<tau> n" succeed full_af 
  by (unfold_locales, rule rpo_pr_prc.RPO_S_pr_ctxt_closed)


(* More efficient implementation which does not require the bound *)

function rpo_unbounded :: "('f \<Rightarrow> nat \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> bool \<times> bool) \<times> ('f \<Rightarrow> bool) \<Rightarrow> ('f \<Rightarrow> nat \<Rightarrow> order_tag) \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> bool \<times> bool" where
  "rpo_unbounded _ _ (Var x) (Var y) = (False, x = y)"
| "rpo_unbounded pr _ (Var x) (Fun g ts) = (False, ts = [] \<and> snd pr g)"
| "rpo_unbounded pr c (Fun f ss) (Var y) =
    (let con = \<exists>s \<in> set ss. snd (rpo_unbounded pr c s (Var y)) in (con,con))" 
| "rpo_unbounded pr c (Fun f ss) (Fun g ts) = (
    if \<exists>s \<in> set ss. snd (rpo_unbounded pr c s (Fun g ts)) 
    then (True,True)
    else (let (prs,prns) = (fst pr) f (length ss) g (length ts) in
         if prns \<and> (\<forall>t \<in> set ts. fst (rpo_unbounded pr c (Fun f ss) t))
         then if prs
              then (True,True) 
              else if c f (length ss) = c g (length ts)
                   then if c f (length ss) = Mul
                        then mul_ext (rpo_unbounded pr c) ss ts
                        else lex_ext_unbounded (rpo_unbounded pr c) ss ts
                   else (length ss \<noteq> 0 \<and> length ts = 0, length ts = 0)
         else (False,False)))"
by pat_completeness auto

termination rpo_unbounded
proof (relation "measures [\<lambda> (pr,c,s,t). size s + size t]")
{
  fix "pr" c and f :: 'f and ss :: "('f,'v)term list" and g :: 'f and ts :: "('f,'v)term list" and x xa y xb x'
  assume s: "xb \<in> set ss" and t: "x' \<in> set ts"
  from s have s1: "size xb \<le> list_size size ss" by (induct ss, auto)
  from t have t1: "size x' \<le> list_size size ts" by (induct ts, auto)
  from s1 t1 s t show "((pr, c, xb, x'), pr, c, Fun f ss, Fun g ts) \<in> measures [\<lambda> (pr,c,s,t). size s + size t]"
    by auto
}
moreover
{
  fix "pr" c and f :: 'f and ss :: "('f,'v)term list" and g :: 'f and ts :: "('f,'v)term list" and x xa xb xc i
  assume "i < length ss" and "i < length ts"
  then obtain s t where si: "s = ss ! i" and s: "s \<in> set ss" and ti: "t = ts ! i" and t: "t \<in> set ts" by auto
  from s have s1: "size s \<le> list_size size ss" by (induct ss, auto)
  from t have t1: "size t \<le> list_size size ts" by (induct ts, auto)
  from s1 t1 si ti show "((pr, c, ss ! i, ts ! i), pr, c, Fun f ss, Fun g ts) \<in> measures [\<lambda> (pr,c,s,t). size s + size t]"
    by auto
}
qed (auto simp: size_simps)

lemma rpo_unbounded_stri_imp_nstri[rule_format]: "fst (rpo_unbounded pr c s t) \<longrightarrow>
  snd (rpo_unbounded pr c s t)"
proof (induct rule: rpo_unbounded.induct[of "\<lambda> pr c s t. fst (rpo_unbounded pr c s t) \<longrightarrow> snd (rpo_unbounded pr c s t)"])
  case (4 "pr" c f ss g ts)
  obtain prc prl where prec: "pr = (prc,prl)"  by (cases "pr", auto)
  obtain  s ns where  prc: "prc f (length ss) g (length ts) = (s,ns)"
   by (cases "prc f (length ss) g (length ts)", auto)
  show ?case
    by (auto simp: Let_def lex_ext_unbounded_stri_imp_nstri
      mul_ext_stri_imp_nstri prec prc)
qed (auto simp: Let_def)

(* Using these unbounded version is harmless *)

lemma rpo_to_rpo_unbounded:
  assumes "\<forall> f i. (f, i) \<in> funas_term s \<union> funas_term t \<longrightarrow> i \<le> n" (is "?b s t")
  shows "rpo pr c n s t = rpo_unbounded pr c s t" (is "?e pr c s t")
proof -
let ?p = "\<lambda>p c s t. ?b s t \<longrightarrow> ?e p c s t"
{
  fix p c s t
  have "?p p c s t"
  proof (induct rule: rpo_unbounded.induct[where P = ?p])
  case (1 p c x y) show ?case by simp
  next
  case (2 p c x g ts) show ?case by simp
  next
  case (3 p c f ss y)
    show ?case 
    proof (intro impI)
      assume "?b (Fun f ss) (Var y)"
      hence "\<And> s. s \<in> set ss \<Longrightarrow> ?b s (Var y)" by auto
      with 3 show "?e p c (Fun f ss) (Var y)" by simp
    qed
  next
  case (4 p c f ss g ts) note IH = this
    show ?case
    proof (intro impI)
    assume "?b (Fun f ss) (Fun g ts)"
    hence bs: "\<And> s. s \<in> set ss \<Longrightarrow> ?b s (Fun g ts)"
     and bt: "\<And> t. t \<in> set ts \<Longrightarrow> ?b (Fun f ss) t"
     and ss: "length ss \<le> n" and ts: "length ts \<le> n" by auto
    with 4(1) have s: "\<And> s. s \<in> set ss \<Longrightarrow> ?e p c s (Fun g ts)" by simp
    show "?e p c (Fun f ss) (Fun g ts)"
    proof (cases "\<exists> s \<in> set ss. snd (rpo p c n s (Fun g ts))")
    case True with s show ?thesis by simp
    next
    case False note oFalse = this
      with s have oFalse2: "\<not> (\<exists>s \<in> set ss. snd (rpo_unbounded p c s (Fun g ts)))"
       by simp
      obtain prns prs where Hsns: "(fst p) f (length ss) g (length ts) = (prs, prns)"
       by (cases "(fst p) f (length ss) g (length ts)") auto
      with bt 4(2)[OF oFalse2 _ Hsns[symmetric]]
       have t: "\<And> t. t \<in> set ts \<Longrightarrow> ?e p c (Fun f ss) t" by force
      show ?thesis
      proof (cases "prns \<and> (\<forall>t\<in>set ts. fst (rpo p c n (Fun f ss) t))")
      case False
        show ?thesis
        proof (cases "prns")
        case False thus ?thesis by (simp add: oFalse oFalse2 Hsns)
        next
        case True
          with False have Hf1: "\<not> (\<forall>t\<in>set ts. fst (rpo p c n (Fun f ss) t))" by simp
          with t have Hf2: "\<not> (\<forall>t\<in>set ts. fst (rpo_unbounded p c (Fun f ss) t))" by auto
          show ?thesis by (simp add: oFalse oFalse2 Hf1 Hf2)
        qed         
      next
      case True
        hence prns: "prns" and Hts: "\<forall>t\<in>set ts. fst (rpo p c n (Fun f ss) t)" by auto
        from Hts and t have Hts2: "\<forall>t\<in>set ts. fst (rpo_unbounded p c (Fun f ss) t)" by auto
        show ?thesis
        proof (cases "prs")
        case True thus ?thesis by (simp add: oFalse oFalse2 Hsns prns Hts Hts2)
        next
        case False note prs = this
          show ?thesis
          proof (cases "c f (length ss) = c g (length ts)")
          case False thus ?thesis
            by (cases "c f (length ss)", simp_all add: oFalse oFalse2 Hsns prns Hts Hts2)
          next
          case True note cfg = this
            show ?thesis
            proof (cases "c f (length ss)")
            case Mul note cf = this
              with cfg have cg: "c g (length ts) = Mul" by simp
              {
                fix x y
                assume x_in_ss: "x \<in> set ss" and y_in_ts: "y \<in> set ts"
                from 4(3)[OF oFalse2 _ Hsns[symmetric] _ prs cfg cf
                 x_in_ss y_in_ts] prns Hts2 bs[OF x_in_ss] bt[OF y_in_ts]
                 have "rpo p c n x y = rpo_unbounded p c x y" by auto
              }
              with mul_ext_cong[of ss ss ts ts]
               have "mul_ext (rpo p c n) ss ts = mul_ext (rpo_unbounded p c) ss ts"
               by metis
              thus ?thesis
               by (simp add: oFalse oFalse2 Hsns prns Hts Hts2 cf cg)
            next
            case Lex note cf = this
              hence ncf: "c f (length ss) \<noteq> Mul" by simp
              from cf cfg have cg: "c g (length ts) = Lex" by simp
              {
                fix i
                assume iss: "i < length ss" and its: "i < length ts"
                from nth_mem_multiset_of[OF iss] and in_multiset_in_set
                 have in_ss: "ss ! i \<in> set ss" by force
                from nth_mem_multiset_of[OF its] and in_multiset_in_set
                 have in_ts: "ts ! i \<in> set ts" by force
                from 4(4)[OF oFalse2 _ Hsns[symmetric] _ prs cfg ncf iss its]
                 prns Hts2 bs[OF in_ss] bt[OF in_ts]
                 have "rpo p c n (ss ! i) (ts ! i) = rpo_unbounded p c (ss ! i) (ts ! i)"
                 by simp
              }
              with lex_ext_cong[of ss ss n n ts ts]
               have "lex_ext (rpo p c n) n ss ts
                = lex_ext (rpo_unbounded p c) n ss ts" by metis
              hence " lex_ext (rpo p c n) n ss ts = lex_ext_unbounded (rpo_unbounded p c) ss ts"
               by (simp add: lex_ext_to_lex_ext_unbounded[OF ss ts, of "rpo_unbounded p c"])
              thus ?thesis
               by (simp add: oFalse oFalse2 Hsns prns prs Hts Hts2 cf cg)
            qed
          qed
        qed
      qed
    qed
    qed
  qed
}
thus ?thesis using assms by simp
qed

definition
  rpo_strict_unbounded
where
  "rpo_strict_unbounded pr c \<equiv> \<lambda>(s, t). check (fst (rpo_unbounded pr c s t)) (''could not orient '' +#+ shows s +@+ '' >RPO '' +#+ shows t +@+ shows_nl)"

definition
  rpo_nstrict_unbounded
where
  "rpo_nstrict_unbounded pr c \<equiv> \<lambda>(s, t). check (snd (rpo_unbounded pr c s t)) (''could not orient '' +#+ shows s +@+ '' >=RPO '' +#+ shows t +@+ shows_nl)"

type_synonym 'f status_prec_repr = "(('f \<times> nat) \<times> (nat \<times> order_tag))list"

fun shows_rpo_repr :: "('f::show)status_prec_repr \<Rightarrow> shows"
where "shows_rpo_repr prs = (''RPO with the following precedence'' +#+ shows_nl +@+ shows_map (\<lambda>((f,n),(pr,s)).
  ''precedence('' +#+ shows f +@+ ''['' +#+ shows n +@+ '']) = '' +#+ shows pr +@+ shows_nl
) prs +@+ shows_nl
+@+ ''precedence(_) = 0'' +#+ shows_nl +@+
  ''and the following status'' +#+ shows_nl +@+
shows_map (\<lambda>((f,n),(pr,s)).
  ''status('' +#+ shows f +@+ ''['' +#+ shows n +@+ '']) = '' +#+ (case s of Mul \<Rightarrow> ''mul'' | Lex \<Rightarrow> ''lex'') +#+ shows_nl
) prs +@+ shows_nl
+@+ ''status(_) = lex'' +#+ shows_nl
)"

fun create_RPO_redtriple :: "(('f :: show) status_prec_repr \<Rightarrow> ('g \<Rightarrow> nat \<Rightarrow> nat) \<times> ('g \<Rightarrow> nat \<Rightarrow> order_tag)) \<Rightarrow> 'f status_prec_repr \<Rightarrow> ('g :: show,'v :: show)redtriple"
where "create_RPO_redtriple prec_repr_to_pr pr = (let (p,\<tau>) = prec_repr_to_pr pr;
   ns = rpo_nstrict_unbounded (pr_nat p) \<tau> in
  \<lparr> redtriple.valid = succeed, s = rpo_strict_unbounded (pr_nat p) \<tau>, ns = ns, nst = ns, af = full_af, mono = (\<lambda> _. succeed), desc =shows_rpo_repr pr, cpx = no_complexity\<rparr>)"

interpretation RPO_redpair: generic_redtriple_impl "create_RPO_redtriple (prec_repr_to_pr :: ('f :: show)status_prec_repr \<Rightarrow> ('g :: show \<Rightarrow> nat \<Rightarrow> nat) \<times> ('g \<Rightarrow> nat \<Rightarrow> order_tag))"
proof
  fix "pr\<tau>" :: "('f :: show)status_prec_repr"
   and s_list ns_list nst_list :: "('g :: show,'v :: show)rule list"
  let ?rp = "create_RPO_redtriple prec_repr_to_pr pr\<tau> :: ('g,'v)redtriple"
  let ?af = "redtriple.af ?rp :: ('g af)"
  let ?cpx = "redtriple.cpx ?rp"
  let ?pr = "prec_repr_to_pr pr\<tau>"
  obtain "pr" \<tau> where id: "?pr = (pr,\<tau>)" by (cases ?pr, auto)
  assume stri: "isOK(check_allm (redtriple.s ?rp) s_list)"
   and nstri: "isOK (check_allm (redtriple.ns ?rp) ns_list)"
   and nstri_top: "isOK (check_allm (redtriple.nst ?rp) nst_list)"
  obtain n where n: "n = max_list (map (\<lambda> (s,t). max_list (map snd ((funas_term_list s) @ (funas_term_list t)))) (s_list @ ns_list @ nst_list))" by auto
  {
    fix s t
    assume st: "(s,t) \<in> set (s_list @ ns_list @ nst_list)"
    have "\<forall> f i. (f,i) \<in> funas_term s \<union> funas_term t \<longrightarrow> i \<le> n"
    proof (intro allI impI)
      fix f i
      assume fi: "(f,i) \<in> funas_term s \<union> funas_term t"
      from max_list st have one: "max_list (map snd ((funas_term_list s) @ (funas_term_list t))) \<le> n" (is "?ml \<le> n") unfolding n by force
      from fi max_list have two: "i \<le> ?ml" by force
      from one two show "i \<le> n" by simp
    qed
    hence "rpo (pr_nat pr) \<tau> n s t = rpo_unbounded (pr_nat pr) \<tau> s t"
     by (rule rpo_to_rpo_unbounded)
  } note main = this
  have mono: "mono_ce_af_redtriple_order (RPO_S pr \<tau> n) (RPO_NS pr \<tau> n) (RPO_NS pr \<tau> n) full_af" ..
  have ce_af: "cpx_ce_af_redtriple_order (RPO_S pr \<tau> n) (RPO_NS pr \<tau> n) (RPO_NS pr \<tau> n) full_af no_complexity" ..
  show "\<exists> S NS NST. cpx_ce_af_redtriple_order S NS NST ?af ?cpx \<and> set s_list \<subseteq> S \<and> set ns_list \<subseteq> NS \<and> set nst_list \<subseteq> NST \<and> (isOK (redtriple.mono ?rp (s_list @ ns_list @ nst_list)) \<longrightarrow> mono_ce_af_redtriple_order S NS NST ?af \<and> ctxt_closed S)"
  proof (simp add: id Let_def, intro exI, intro conjI, rule ce_af) 
    show "ctxt_closed (RPO_S pr \<tau> n :: ('g, 'v) term rel)"
      by (rule RPO_ce_mono_redpair.checkMonoSound, simp)
  next
    {
      fix s t
      assume st: "(s,t) \<in> set s_list"
      with stri have "fst (rpo_unbounded (pr_nat pr) \<tau> s t)"
       by (auto simp: id rpo_strict_unbounded_def Let_def)
      with st have "fst (rpo (pr_nat pr) \<tau> n s t)" by (simp add: main)
      hence "(s,t) \<in> RPO_S pr \<tau> n" unfolding rpo_pr_prc.RPO_S_pr_def by simp
    }
    thus "set s_list \<subseteq> RPO_S pr \<tau> n" ..
  next
    {
      fix s t
      assume st: "(s,t) \<in> set ns_list"
      with nstri have "snd (rpo_unbounded (pr_nat pr) \<tau> s t)"
       by (auto simp: id rpo_nstrict_unbounded_def Let_def)
      with st have "snd (rpo (pr_nat pr) \<tau> n s t)" by (simp add: main)
      hence "(s,t) \<in> RPO_NS pr \<tau> n" unfolding rpo_pr_prc.RPO_NS_pr_def by simp
    }
    thus "set ns_list \<subseteq> RPO_NS pr \<tau> n" ..
  next
    {
      fix s t
      assume st: "(s,t) \<in> set nst_list"
      with nstri_top have "snd (rpo_unbounded (pr_nat pr) \<tau> s t)"
       by (auto simp: id rpo_nstrict_unbounded_def Let_def)
      with st have "snd (rpo (pr_nat pr) \<tau> n s t)" by (simp add: main)
      hence "(s,t) \<in> RPO_NS pr \<tau> n" unfolding rpo_pr_prc.RPO_NS_pr_def by simp
    }
    thus "set nst_list \<subseteq> RPO_NS pr \<tau> n" ..    
  qed (rule mono)
qed

end

